# coding: utf-8
# Copyright (c) Pymatgen Development Team.
# Distributed under the terms of the MIT License.

from __future__ import division, print_function, unicode_literals
from __future__ import absolute_import

"""
This module provides a base class for tensor-like objects and methods for
basic tensor manipulation.  It also provides a class, SquareTensor,
that provides basic methods for creating and manipulating rank 2 tensors
"""


__author__ = "Maarten de Jong"
__copyright__ = "Copyright 2012, The Materials Project"
__credits__ = ("Joseph Montoya, Shyam Dwaraknath, Wei Chen, "
               "Mark Asta, Anubhav Jain, Terence Lew")
__version__ = "1.0"
__maintainer__ = "Joseph Montoya"
__email__ = "montoyjh@lbl.gov"
__status__ = "Development"
__date__ = "March 22, 2012"


from scipy.linalg import polar
from scipy.linalg import sqrtm
import numpy as np
import itertools
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from pymatgen.core.operations import SymmOp
from pymatgen.core.lattice import Lattice
from numpy.linalg import norm

class TensorBase(np.ndarray):
    """
    Base class for doing useful general operations on Nth order tensors,
    without restrictions on the type (stress, elastic, strain, piezo, etc.)
    """

    def __new__(cls, input_array):
        """
        Create a TensorBase object.  Note that the constructor uses __new__
        rather than __init__ according to the standard method of
        subclassing numpy ndarrays.

        Args:
            input_array: (3xN array-like): the 3xN array-like representing
                a tensor quantity
        """
        obj = np.asarray(input_array).view(cls)
        obj.rank = len(obj.shape)
        if not all([i == 3 for i in obj.shape]):
            raise ValueError("Pymatgen only supports 3-dimensional tensors")
        
        return obj

    def __array_finalize__(self, obj):
        if obj is None:
            return
        self.rank = getattr(obj, 'rank', None)

    def __array_wrap__(self, obj):
        """
        Overrides __array_wrap__ methods in ndarray superclass to avoid errors
        associated with functions that return scalar values
        """

        if len(obj.shape) == 0:
            return obj[()]
        else:
            return np.ndarray.__array_wrap__(self, obj)
        
    def __hash__(self):
        """
        define a hash function, since numpy arrays
        have their own __eq__ method
        """
        return hash(self.tostring())

    def __repr__(cls):
        return "{}({})".format(cls.__class__.__name__,
                               cls.__str__())

    def zeroed(self, tol = 1e-3):
        """
        returns the matrix with all entries below a certain threshold
        (i.e. tol) set to zero
        """
        new_tensor = self.copy()
        new_tensor[abs(new_tensor) < tol] = 0
        return new_tensor

    def transform(self, symm_op):
        """
        Applies a transformation (via a symmetry operation) to a tensor. 

        Args:
            symm_op (SymmOp): a symmetry operation to apply to the tensor
        """
        return self.__class__(symm_op.transform_tensor(self))

    def rotate(self, matrix, tol=1e-3):
        """
        Applies a rotation directly, and tests input matrix to ensure a valid
        rotation.

        Args:
            matrix (3x3 array-like): rotation matrix to be applied to tensor
            tol (float): tolerance for testing rotation matrix validity
        """
        matrix = SquareTensor(matrix)
        if not matrix.is_rotation(tol):
            raise ValueError("Rotation matrix is not valid.")
        sop = SymmOp.from_rotation_and_translation(matrix,
                                                   [0., 0., 0.])
        return self.transform(sop)

    @property
    def symmetrized(self):
        """
        Returns a generally symmetrized tensor, calculated by taking 
        the sum of the tensor and its transpose with respect to all 
        possible permutations of indices
        """
        perms = list(itertools.permutations(range(self.rank)))
        return sum([np.transpose(self, ind) for ind in perms]) / len(perms)

    def is_symmetric(self, tol=1e-5):
        """
        Tests whether a tensor is symmetric or not based on the residual
        with its symmetric part, from self.symmetrized

        Args:
            tol (float): tolerance to test for symmetry
        """
        return (self - self.symmetrized < tol).all()

    def fit_to_structure(self, structure, symprec = 0.1):
        """
        Returns a tensor that is invariant with respect to symmetry
        operations corresponding to a structure

        Args: 
            structure (Structure): structure from which to generate 
                symmetry operations
            symprec (float): symmetry tolerance for the Spacegroup Analyzer
                used to generate the symmetry operations
        """
        sga = SpacegroupAnalyzer(structure, symprec)
        symm_ops = sga.get_symmetry_operations(cartesian=True)
        return sum([self.transform(symm_op)
                    for symm_op in symm_ops]) / len(symm_ops)

    def is_fit_to_structure(self, structure, tol=1e-2):
        """
        Tests whether a tensor is invariant with respect to the
        symmetry operations of a particular structure by testing
        whether the residual of the symmetric portion is below a 
        tolerance
        
        Args:
            tol (float): tolerance for symmetry testing
        """
        return (self - self.fit_to_structure(structure) < tol).all()

    def convert_to_ieee(self, structure, atol = 0.1, ltol = 0.05):
        """
        Given a structure associated with a tensor, attempts a
        calculation of the tensor in IEEE format according to
        the 1987 IEEE standards.

        Args:
            structure (Structure): a structure associated with the
                tensor to be converted to the IEEE standard
            atol (float): angle tolerance for conversion routines
            ltol (float): length tolerance for conversion routines
        """
        def get_uvec(vec):
            """ Gets a unit vector parallel to input vector"""
            return vec / np.linalg.norm(vec)

        # Check conventional setting:
        sga = SpacegroupAnalyzer(structure)
        dataset = sga.get_symmetry_dataset()
        trans_mat = dataset['transformation_matrix']
        conv_latt = Lattice(np.transpose(np.dot(np.transpose(
            structure.lattice.matrix), np.linalg.inv(trans_mat))))
        xtal_sys = sga.get_crystal_system()
        
        vecs = conv_latt.matrix
        lengths = np.array(conv_latt.abc)
        angles = np.array(conv_latt.angles)
        a = b = c = None
        rotation = np.zeros((3,3))

        # IEEE rules: a,b,c || x1,x2,x3
        if xtal_sys == "cubic":
            rotation = [vecs[i]/lengths[i] for i in range(3)]

        # IEEE rules: a=b in length; c,a || x3, x1
        elif xtal_sys == "tetragonal":
            rotation = np.array([vec/mag for (mag, vec) in 
                                 sorted(zip(lengths, vecs))])
            if abs(lengths[2] - lengths[1]) < ltol:
                rotation[0], rotation[2] = rotation[2], rotation[0].copy()
            rotation[1] = get_uvec(np.cross(rotation[2], rotation[0]))

        # IEEE rules: c<a<b; c,a || x3,x1
        elif xtal_sys == "orthorhombic":
            rotation = [vec/mag for (mag, vec) in sorted(zip(lengths, vecs))]
            rotation = np.roll(rotation, 2, axis = 0)

        # IEEE rules: c,a || x3,x1, c is threefold axis
        # Note this also includes rhombohedral crystal systems
        elif xtal_sys in ("trigonal", "hexagonal"):
            # find threefold axis:
            tf_mask = abs(angles-120.0) < atol
            non_tf_mask = np.logical_not(tf_mask)
            rotation[2] = get_uvec(vecs[tf_mask][0])
            rotation[0] = get_uvec(vecs[non_tf_mask][0])
            rotation[1] = get_uvec(np.cross(rotation[2], rotation[0]))

        # IEEE rules: b,c || x2,x3; alpha=beta=90, c<a
        elif xtal_sys == "monoclinic":
            # Find unique axis
            umask = abs(angles - 90.0) > atol
            n_umask = np.logical_not(umask)
            rotation[1] = get_uvec(vecs[umask])
            # Shorter of remaining lattice vectors for c axis
            c = [vec/mag for (mag, vec) in 
                 sorted(zip(lengths[n_umask], vecs[n_umask]))][0]
            rotation[2] = np.array(c)
            rotation[0] = np.cross(rotation[1], rotation[2])
        
        # IEEE rules: c || x3
        elif xtal_sys == "triclinic":
            rotation = [vec/mag for (mag, vec) in sorted(zip(lengths, vecs))]
            rotation = np.roll(rotation, 2, axis = 0)
            rotation[1] = get_uvec(np.cross(rotation[2], rotation[1]))
            rotation[0] = np.cross(rotation[1], rotation[2])
        
        return self.rotate(rotation)


class SquareTensor(TensorBase):
    """
    Base class for doing useful general operations on second rank tensors
    (stress, strain etc.).
    """

    def __new__(cls, input_array):
        """
        Create a SquareTensor object.  Note that the constructor uses __new__
        rather than __init__ according to the standard method of
        subclassing numpy ndarrays.  Error is thrown when the class is
        initialized with non-square matrix.

        Args:
            stress_matrix (3x3 array-like): the 3x3 array-like
                representing the Green-Lagrange strain
        """

        obj = TensorBase(input_array).view(cls)
        if not (len(obj.shape) == 2):
            raise ValueError("SquareTensor only takes 2-D "
                             "tensors as input")
        return obj
        
    @property
    def trans(self):
        """
        shorthand for transpose on SquareTensor
        """
        return SquareTensor(np.transpose(self))

    @property
    def inv(self):
        """
        shorthand for matrix inverse on SquareTensor
        """
        if self.det == 0:
            raise ValueError("SquareTensor is non-invertible")
        return SquareTensor(np.linalg.inv(self))

    @property
    def det(self):
        """
        shorthand for the determinant of the SquareTensor
        """
        return np.linalg.det(self)

    def is_rotation(self, tol=1e-3):
        """
        Test to see if tensor is a valid rotation matrix, performs a
        test to check whether the inverse is equal to the transpose
        and if the determinant is equal to one within the specified
        tolerance

        Args:
            tol (float): tolerance to both tests of whether the
                the determinant is one and the inverse is equal
                to the transpose
        """

        return (np.abs(self.inv - self.trans) < tol).all() \
            and (np.linalg.det(self) - 1. < tol)

    def get_scaled(self, scale_factor):
        """
        Scales the tensor by a certain multiplicative scale factor

        Args:
            scale_factor (float): scalar multiplier to be applied to the
                SquareTensor object
        """
        return SquareTensor(self * scale_factor)

    @property
    def principal_invariants(self):
        """
        Returns a list of principal invariants for the tensor,
        which are the values of the coefficients of the characteristic
        polynomial for the matrix
        """
        return np.poly(self)[1:]*np.array([-1, 1, -1])

    def polar_decomposition(self, side='right'):
        """
        calculates matrices for polar decomposition
        """
        return polar(self, side=side)



# coding: utf-8
# Copyright (c) Pymatgen Development Team.
# Distributed under the terms of the MIT License.

from __future__ import division, unicode_literals

"""
This module defines classes for point defects
"""

import os
import abc
import json
from bisect import bisect_left

from pymatgen.core.periodic_table import Specie, Element
from pymatgen.core.sites import PeriodicSite
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from pymatgen.io.zeopp import get_voronoi_nodes, get_void_volume_surfarea, \
    get_high_accuracy_voronoi_nodes
from pymatgen.command_line.gulp_caller import get_energy_buckingham, \
    get_energy_relax_structure_buckingham
from pymatgen.analysis.structure_analyzer import VoronoiCoordFinder, \
    RelaxationAnalyzer
from pymatgen.analysis.structure_matcher import StructureMatcher
from pymatgen.analysis.bond_valence import BVAnalyzer
import six
from six.moves import filter
from six.moves import map
from six.moves import zip

file_dir = os.path.dirname(__file__)
rad_file = os.path.join(file_dir, 'ionic_radii.json')
with open(rad_file, 'r') as fp:
    _ion_radii = json.load(fp)


class ValenceIonicRadiusEvaluator(object):
    """
    Computes site valences and ionic radii for a structure using bond valence
    analyzer

    Args:
        structure: pymatgen.core.structure.Structure
    """

    def __init__(self, structure):
        self._structure = structure.copy()
        self._valences = self._get_valences()
        self._ionic_radii = self._get_ionic_radii()

    @property
    def radii(self):
        """
        List of ionic radii of elements in the order of sites.
        """
        el = [site.species_string for site in self._structure.sites]
        radii_dict = dict(zip(el, self._ionic_radii))
        #print radii_dict
        return radii_dict

    @property
    def valences(self):
        """
        List of oxidation states of elements in the order of sites.
        """
        el = [site.species_string for site in self._structure.sites]
        valence_dict = dict(zip(el, self._valences))
        return valence_dict

    @property
    def structure(self):
        """
        Returns oxidation state decorated structurel.
        """
        return self._structure.copy()


    def _get_ionic_radii(self):
        """
        Computes ionic radii of elements for all sites in the structure.
        If valence is zero, atomic radius is used.
        """
        radii = []
        coord_finder = VoronoiCoordFinder(self._structure)

        def nearest_key(sorted_vals, key):
            i = bisect_left(sorted_vals, key)
            if i == len(sorted_vals):
                return sorted_vals[-1]
            if i == 0:
                return sorted_vals[0]
            before = sorted_vals[i-1]
            after = sorted_vals[i]
            if after-key < key-before:
                return after
            else:
                return before

        for i in range(len(self._structure.sites)):
            site = self._structure.sites[i]
            if isinstance(site.specie,Element):
                radius = site.specie.atomic_radius
                # Handle elements with no atomic_radius
                # by using calculated values instead.
                if radius == None:
                    radius = site.specie.atomic_radius_calculated
                if radius == None:
                    raise ValueError(
                            "cannot assign radius to element {}".format(
                            site.specie))
                radii.append(radius)
                continue

            el = site.specie.symbol
            oxi_state = int(round(site.specie.oxi_state))
            coord_no = int(round(coord_finder.get_coordination_number(i)))
            try:
                tab_oxi_states = sorted(map(int, _ion_radii[el].keys()))
                oxi_state = nearest_key(tab_oxi_states, oxi_state)
                radius = _ion_radii[el][str(oxi_state)][str(coord_no)]
            except KeyError:
                if coord_finder.get_coordination_number(i)-coord_no > 0:
                    new_coord_no = coord_no + 1
                else:
                    new_coord_no = coord_no - 1
                try:
                    radius = _ion_radii[el][str(oxi_state)][str(new_coord_no)]
                    coord_no = new_coord_no
                except:
                    tab_coords = sorted(map(int, _ion_radii[el][str(oxi_state)].keys()))
                    new_coord_no = nearest_key(tab_coords, coord_no)
                    i = 0
                    for val in tab_coords:
                        if  val > coord_no:
                            break
                        i = i + 1
                    if i == len(tab_coords):
                        key = str(tab_coords[-1])
                        radius = _ion_radii[el][str(oxi_state)][key]
                    elif i == 0:
                        key = str(tab_coords[0])
                        radius = _ion_radii[el][str(oxi_state)][key]
                    else:
                        key = str(tab_coords[i-1])
                        radius1 = _ion_radii[el][str(oxi_state)][key]
                        key = str(tab_coords[i])
                        radius2 = _ion_radii[el][str(oxi_state)][key]
                        radius = (radius1+radius2)/2

            #implement complex checks later
            radii.append(radius)
        return radii

    def _get_valences(self):
        """
        Computes ionic valences of elements for all sites in the structure.
        """
        try:
            bv = BVAnalyzer()
            self._structure = bv.get_oxi_state_decorated_structure(self._structure)
            valences = bv.get_valences(self._structure)
        except:
            try:
                bv = BVAnalyzer(symm_tol=0.0)
                self._structure = bv.get_oxi_state_decorated_structure(self._structure)
                valences = bv.get_valences(self._structure)
            except:
                valences = []
                for site in self._structure.sites:
                    if len(site.specie.common_oxidation_states) > 0:
                        valences.append(site.specie.common_oxidation_states[0])
                    # Handle noble gas species
                    # which have no entries in common_oxidation_states.
                    else:
                        valences.append(0)
                if sum(valences):
                    valences = [0]*self._structure.num_sites
                else:
                    self._structure.add_oxidation_state_by_site(valences)
                #raise

        #el = [site.specie.symbol for site in self._structure.sites]
        #el = [site.species_string for site in self._structure.sites]
        #el = [site.specie for site in self._structure.sites]
        #valence_dict = dict(zip(el, valences))
        #print valence_dict
        return valences


class Defect(six.with_metaclass(abc.ABCMeta, object)):
    """
    Abstract class for point defects
    """

    @abc.abstractmethod
    def enumerate_defectsites(self):
        """
        Enumerates all the symmetrically distinct defects.
        """
        raise NotImplementedError()

    @property
    def structure(self):
        """
        Returns the structure without any defects
        Useful for Mott-Littleton calculations.
        """
        return self._structure

    @property
    def struct_radii(self):
        """
        Radii of elements in the structure
        """
        return self._rad_dict

    @property
    def struct_valences(self):
        """
        Valence of elements in the structure
        """
        return self._valence_dict

    def defectsite_count(self):
        """
        Returns the number of symmetrically distinct defect sites
        """
        return len(self._defect_sites)

    def get_defectsite(self, n):
        """
        Returns the defect site at the index.
        """
        return self._defect_sites[n]

    def get_defectsite_multiplicity(self, n):
        """
        Returns the symmtric multiplicity of the defect site at the index.
        """
        return self._defect_site_multiplicity[n]

    def get_defectsite_coordination_number(self, n):
        """
        Coordination number of interstitial site.

        Args:
            n: Index of interstitial list
        """
        return self._defectsite_coord_no[n]

    def get_coordinated_sites(self, n):
        """
        The sites in structure surrounding the defect site.

        Args:
            n: Index of defects list
        """
        return self._defect_coord_sites[n]

    def get_coordinated_elements(self, n):
        """
        Elements of sites in structure surrounding the defect site.

        Args:
            n: Index of defect list
        """
        coordinated_species = []
        for site in self._defect_coord_sites[n]:
            coordinated_species.append(site.specie.symbol)
        return list(set(coordinated_species))

    @abc.abstractmethod
    def make_supercells_with_defects(self, scaling_matrix):
        """
        Generate the supercell with input multipliers and create the defect.
        First supercell has no defects.
        To create unit cell with defect pass unit matrix.
        """
        raise NotImplementedError()


class Vacancy(Defect):
    """
    Subclass of Defect to generate vacancies and their analysis.

    Args:
        structure: pymatgen.core.structure.Structure
        valences: valences of elements as a dictionary
        radii: Radii of elements as a dictionary
    """

    def __init__(self, structure, valences, radii):
        self._structure = structure
        self._valence_dict = valences
        self._rad_dict = radii
        # Store symmetrically distinct sites, their coordination numbers
        # coordinated_sites, effective charge
        symm_finder = SpacegroupAnalyzer(self._structure)
        symm_structure = symm_finder.get_symmetrized_structure()
        equiv_site_seq = symm_structure.equivalent_sites

        self._defect_sites = []
        self._defect_site_multiplicity = []
        for equiv_sites in equiv_site_seq:
            self._defect_sites.append(equiv_sites[0])
            self._defect_site_multiplicity.append(len(equiv_sites))

        self._vac_site_indices = []
        for site in self._defect_sites:
            for i in range(len(self._structure.sites)):
                if site == self._structure[i]:
                    self._vac_site_indices.append(i)

        coord_finder = VoronoiCoordFinder(self._structure)
        self._defectsite_coord_no = []
        self._defect_coord_sites = []
        for i in self._vac_site_indices:
            self._defectsite_coord_no.append(
                coord_finder.get_coordination_number(i)
            )
            self._defect_coord_sites.append(
                coord_finder.get_coordinated_sites(i)
            )

        # Store the ionic radii for the elements in the structure
        # (Used to  computing the surface are and volume)
        # Computed based on valence of each element

        self._vac_eff_charges = None
        self._vol = None
        self._sa = None

    #@property
    #def valence_dict(self):
    #    return self._valence_dict

    def enumerate_defectsites(self):
        """
        Returns symmetrically distinct vacancy sites
        """
        return self._defect_sites

    def get_defectsite_structure_indices(self):
        """
        Returns indices of symmetrically distinct vacancy sites
        """
        return self._vac_site_indices

    def get_defectsite_structure_index(self, n):
        """
        index of the vacacy site in the structure.sites list

        Args:
            n:
                Index of vacancy list
        """
        return self._vac_site_indices[n]

    def get_defectsite_effective_charge(self, n):
        """
        Effective charge (In Kroger-Vink notation, cation vacancy has
        effectively -ve charge and anion vacancy has +ve charge.)

        Args:
            n: Index of vacancy list

        Returns:
            Effective charnge of defect site
        """
        # Effective charge (In Kroger-Vink notation, cation vacancy has
        # effectively -ve charge and anion vacancy has +ve charge.) Inverse
        # the BVAnalyzer.get_valences result.

        el = self.get_defectsite(n).species_string
        return -self._valence_dict[el]
        #if not self._vac_eff_charges:
        #    self._vac_eff_charges = []
        #    for site in self.enumerate_defectsites():
        #        specie = site.specie.symbol
        #        self._vac_eff_charges.append(-self._valence_dict[specie])
        #return self._vac_eff_charges[n]

    def get_coordsites_min_max_charge(self, n):
        """
        Minimum and maximum charge of sites surrounding the vacancy site.

        Args:
            n: Index of vacancy list
        """
        bv = BVAnalyzer()
        struct_valences = bv.get_valences(self._structure)
        coordinated_site_valences = []

        def _get_index(site):
            for i in range(len(self._structure.sites)):
                if site.is_periodic_image(self._structure.sites[i]):
                    return i
            raise ValueError("Site not found")

        for site in self._defect_coord_sites[n]:
            ind = _get_index(site)
            coordinated_site_valences.append(struct_valences[ind])
        coordinated_site_valences.sort()
        return coordinated_site_valences[0], coordinated_site_valences[-1]

    # deprecated
    def get_volume(self, n):
        """
        Volume of the nth vacancy

        Args:
            n: Index of symmetrically distinct vacancies list

        Returns:
            floating number representing volume of vacancy
        """
        if not self._vol:
            self._vol = []
            self._sa = []
            um = [[1, 0, 0], [0, 1, 0], [0, 0, 1]]
            sc = self.make_supercells_with_defects(um)[1:]
            rad_dict = self.struct_radii
            for i in range(len(sc)):
                vol, sa = get_void_volume_surfarea(sc[i], rad_dict)
                self._vol.append(vol)
                self._sa.append(sa)

        return self._vol[n]

    # deprecated
    def get_surface_area(self, n):
        """
        Surface area of the nth vacancy

        Args:
            n: Index of symmetrically distinct vacancies list

        Returns:
            floating number representing volume of vacancy
        """
        if not self._sa:
            self._vol = []
            self._sa = []
            um = [[1, 0, 0], [0, 1, 0], [0, 0, 1]]
            supercells = self.make_supercells_with_defects(um)[1:]
            rad_dict = self.struct_radii
            for sc in supercells:
                vol, sa = get_void_volume_surfarea(sc, rad_dict)
                self._vol.append(vol)
                self._sa.append(sa)

        return self._sa[n]

    def _supercell_with_defect(self, scaling_matrix, defect_site):
        sc = self._structure.copy()
        sc.make_supercell(scaling_matrix)
        oldf_coords = defect_site.frac_coords
        coords = defect_site.lattice.get_cartesian_coords(oldf_coords)
        newf_coords = sc.lattice.get_fractional_coords(coords)
        sc_defect_site = PeriodicSite(defect_site.species_and_occu, newf_coords,
                                      sc.lattice,
                                      properties=defect_site.properties)
        for i in range(len(sc.sites)):
            #if sc_defect_site == sc.sites[i]:
            if sc_defect_site.distance(sc.sites[i]) < 1e-3:
                del sc[i]
                return sc
        raise ValueError('Something wrong if reached here')

    def make_supercells_with_defects(self, scaling_matrix, species=None,
                                     limit_return_structures=False):
        """
        Generate sequence of supercells in pymatgen.core.structure.Structure
        format, with each supercell containing one vacancy.

        Args:
            scaling_matrix: super cell scale parameters in matrix forms
            species: Species in list format only for which vacancy supercells
                are required. If not specified all the species are considered.
            limit_return_structures: Boolean or positive number
                If number, only that many structures are returned.

        Returns:
            Supercells with vacancies. First supercell has no defects.
        """
        sc_with_vac = []
        sc = self._structure.copy()
        sc.make_supercell(scaling_matrix)
        sc_with_vac.append(sc)

        if not species:
            species = sc.symbol_set
        if not limit_return_structures:
            limit_return_structures = self.defectsite_count()
        for defect_site in self.enumerate_defectsites():
            if len(sc_with_vac) <= limit_return_structures:
                if isinstance(defect_site.specie,Specie):
                    site_specie = defect_site.specie.element.symbol
                elif isinstance(defect_site.specie,Element):
                    site_specie = defect_site.specie.symbol
                else:
                    raise TypeError("site specie is neither Specie nor Element")

                if site_specie in species:
                    sc_with_vac.append(self._supercell_with_defect(
                        scaling_matrix, defect_site))
        return sc_with_vac


class VacancyFormationEnergy(object):
    """
    Using GULP compute the vacancy formation energy.
    Works only for binary metal oxides due to the use of Buckingham Potentials
    """

    def __init__(self, vacancy):
        self._vacancy = vacancy
        self._energies = []

    def get_energy(self, n, tol=0.5):
        """
        Formation Energy for nth symmetrically distinct vacancy.
        """
        #generate defect free structure energy
        if not self._energies:
            no_vac = self._vacancy.defectsite_count()
            prev_energies = [0.0] * no_vac
            tol_flg = [False] * no_vac
            vac_gulp_kw = ('optimise', 'conp', 'qok')
            val_dict = self._vacancy.struct_valences
            for sp in range(2, 6):
                if not (False in tol_flg):
                    #print sp
                    break
                scale_mat = [[sp, 0, 0], [0, sp, 0], [0, 0, sp]]
                sc = self._vacancy.make_supercells_with_defects(scale_mat)
                blk_energy = get_energy_buckingham(sc[0])
                no = len(sc[0].sites)
                #print no
                for i in range(1, no_vac + 1):
                    if not tol_flg[i - 1]:
                        vac_energy = get_energy_buckingham(
                            sc[i], keywords=vac_gulp_kw,
                            valence_dict=val_dict
                        )
                        form_energy = vac_energy - (no - 1) / no * blk_energy
                        if abs(form_energy - prev_energies[i - 1]) < tol:
                            tol_flg[i - 1] = True
                        prev_energies[i - 1] = form_energy

            self._energies = prev_energies
            self._tol_flg = tol_flg

        if not self._tol_flg[n]:
            print("Caution: tolerance not reached for {0} vacancy".format(n))
        return self._energies[n]


class Interstitial(Defect):
    """
    Subclass of Defect to generate interstitial sites
    """

    def __init__(self, structure, valences, radii, site_type='voronoi_vertex',
                 accuracy='Normal', symmetry_flag=True, oxi_state = False):
        """
        Given a structure, generate symmetrically distinct interstitial sites.
        For a non-ionic structure, use oxi_state=True and give atomic radii.

        Args:
            structure: pymatgen.core.structure.Structure
            valences: Dictionary of oxidation states of elements in 
                {el:valence} form
            radii: Radii of elemnts in the structure
            site_type: "voronoi_vertex" uses voronoi nodes
                "voronoi_edgecenter" uses voronoi polyhedra edge centers
                "voronoi_facecenter" uses voronoi polyhedra face centers
                "all" combines vertices, edgecenters and facecenters.
                Default is "voronoi_vertex"
            accuracy: Flag denoting whether to use high accuracy version 
                of Zeo++. Options are "Normal" and "High". Default is normal.
            symmetry_flag: If True, only returns symmetrically distinct sites
            oxi_state: If False, input structure is considered devoid of 
                oxidation-state decoration. And oxi-state for each site is 
                determined. Use True, if input structure is oxi-state 
                decorated. This option is useful when the structure is 
                not electro-neutral after deleting/adding sites. In that
                case oxi-decorate the structure before deleting/adding the
                sites.
        """
        if not oxi_state:
            self._structure = ValenceIonicRadiusEvaluator(structure).structure
        else:
            self._structure = structure

        self._valence_dict = valences
        self._rad_dict = radii

        """
        Use Zeo++ to obtain the voronoi nodes. Apply symmetry reduction
        and the symmetry reduced voronoi nodes are possible candidates
        for interstitial sites.
        """
        if accuracy == "Normal":
            high_accuracy_flag = False
        elif accuracy == "High":
            high_accuracy_flag = True
        else:
            raise NotImplementedError("Accuracy setting not implemented.")

        if accuracy == "High":
            if site_type in ('voronoi_facecenter','voronoi_edgecenter','all'): 
                raise NotImplementedError(
                        "Site type not implemented for the accuracy setting")


        vor_node_sites, vor_edgecenter_sites, vor_facecenter_sites = \
            symmetry_reduced_voronoi_nodes(self._structure, self._rad_dict,
                                           high_accuracy_flag, symmetry_flag)
        
        if site_type == 'voronoi_vertex':
            possible_interstitial_sites = vor_node_sites
        elif site_type == 'voronoi_facecenter':
            possible_interstitial_sites = vor_facecenter_sites
        elif site_type == 'voronoi_edgecenter':
            possible_interstitial_sites = vor_edgecenter_sites
        elif site_type == "all":
            possible_interstitial_sites = vor_node_sites + \
                    vor_facecenter_sites + vor_edgecenter_sites
        else:
            raise ValueError("Input site type not implemented")

        #Do futher processing on possibleInterstitialSites to obtain
        #interstitial sites
        self._defect_sites = possible_interstitial_sites
        self._defectsite_coord_no = []
        self._defect_coord_sites = []
        self._defect_coord_charge = []
        self._radii = []

        for site in self._defect_sites:
            coord_no, coord_sites, chrg = self._get_coord_no_sites_chrg(site)
            self._defectsite_coord_no.append(coord_no)
            self._defect_coord_sites.append(coord_sites)
            self._defect_coord_charge.append(chrg)

        for site in self._defect_sites:
            vor_radius = site.properties.get('voronoi_radius',None)
            if vor_radius:
                vor_radius = float(vor_radius)
            self._radii.append(vor_radius)

    def _get_coord_no_sites_chrg(self, site):
        """
        Compute the coordination number and coordination charge

        Args:
            site:
                pymatgen.core.sites.Site
        """
        struct = self._structure.copy()
        struct.append(site.specie.symbol, site.frac_coords)
        coord_finder = VoronoiCoordFinder(struct)
        coord_no = coord_finder.get_coordination_number(-1)
        coord_sites = coord_finder.get_coordinated_sites(-1)

        # In some cases coordination sites to interstitials include
        # interstitials also. Filtering them.
        def no_inter(site):
            return not site.specie.symbol == 'X'
        coord_sites = filter(no_inter, coord_sites)

        coord_chrg = 0
        for site, weight in coord_finder.get_voronoi_polyhedra(-1).items():
            if not site.specie.symbol == 'X':
                coord_chrg += weight * self._valence_dict[site.species_string]

        return coord_no, coord_sites, coord_chrg

    def enumerate_defectsites(self):
        """
        Enumerate all the symmetrically distinct interstitial sites.
        The defect site has "X" as occupied specie.
        """
        return self._defect_sites

    def append_defectsite(self, site):
        """
        Append a site to list of possible interstitials

        Args:
            site: pymatgen.core.sites.Site
        """
        raise NotImplementedError()

    def delete_defectsite(self, n):
        """
        Remove a symmetrically distinct interstitial site

        Args:
            n: Index of interstitial site
        """
        del self._defect_sites[n]

    def get_coordsites_charge_sum(self, n):
        """
        Total charge of the interstitial coordinated sites.

        Args:
            n: Index of interstitial list
        """
        return self._defect_coord_charge[n]

    def get_coordsites_min_max_charge(self, n):
        """
        Minimum and maximum charge of sites surrounding the interstitial site.

        Args:
            n: Index of symmetrical distinct interstitial site
        """
        coord_site_valences = []

        for site in self._defect_coord_sites[n]:
            coord_site_valences.append(self._valence_dict[site.specie.symbol])
        coord_site_valences.sort()
        return coord_site_valences[0], coord_site_valences[-1]

    def get_radius(self, n):
        """
        Volume of the nth interstitial

        Args:
            n: Index of symmetrically distinct vacancies list

        Returns:
            floating number representing radius of interstitial sphere
        """
        return self._radii[n]

    def get_radii(self):
        return self._radii

    def reduce_defectsites(self):
        """
        If multiple defect sites have same voronoi radius, only one is kept.
        Useful if the symmetry based reduction of initial sites returned
        from Zeo++ is not working properly due to deviation in ideal lattice
        coordinates.
        """
        distinct_radii = list(set(self._radii))
        for rad in distinct_radii:
            ind = self._radii.index(rad)  # Index of first site with 'rad'
            for i in reversed(list(range(ind + 1, len(self._radii)))):
                # Backward search for remaining sites so index is not changed
                if self._radii[i] == rad:
                    self._defect_sites.pop(i)
                    self._defectsite_coord_no.pop(i)
                    self._defect_coord_sites.pop(i)
                    self._radii.pop(i)

    def radius_prune_defectsites(self, radius):
        """
        Remove all the defect sites with voronoi radius less than input radius
        """
        for i in reversed(list(range(len(self._radii)))):
            if self._radii[i] < radius:
                self._defect_sites.pop(i)
                self._defectsite_coord_no.pop(i)
                self._defect_coord_sites.pop(i)
                self._radii.pop(i)

    def prune_defectsites(self, el="C", oxi_state=4, dlta=0.1):
        """
        Prune all the defect sites which can't acoomodate the input elment
        with the input oxidation state.
        """
        rad = float(Specie(el, oxi_state).ionic_radius) - dlta
        self.radius_prune_defectsites(rad)

    def prune_close_defectsites(self, dist=0.2):
        """
        Prune the sites that are very close.
        """
        #print self.defectsite_count()
        ind = 0
        while ind < self.defectsite_count():
            #i = ind + 1
            #while i < self.defectsite_count():
            i = self.defectsite_count()-1
            #print ind, i
            while i > ind: 
                d = self._defect_sites[ind].distance(self._defect_sites[i])
                #print d, dist
                if d < dist:
                    self._defect_sites.pop(i)
                    #self._defectsite_coord_no.pop(i)
                    #self._defect_coord_sites.pop(i)
                    #self._radii.pop(i)
            #    i += 1
                i -= 1
            ind += 1
        #print self.defectsite_count()

    def _supercell_with_defect(self, scaling_matrix, defect_site, element):
        sc = self._structure.copy()
        sc.make_supercell(scaling_matrix)
        oldf_coords = defect_site.frac_coords
        coords = defect_site.lattice.get_cartesian_coords(oldf_coords)
        #print coords
        newf_coords = sc.lattice.get_fractional_coords(coords)
        for i in range(3):
            coord = newf_coords[i]
            if coord < 0:
                while (coord < 0):
                    coord = coord+1
                newf_coords[i] = coord
            elif coord > 1:
                while (coord > 1):
                    coord = coord-1
                newf_coords[i] = coord
        #print newf_coords
        #sc_defect_site = PeriodicSite(element, newf_coords,
        #                              sc.lattice)
        try:
            sc.append(element, newf_coords, coords_are_cartesian=False,
                      validate_proximity=True)
        except ValueError:
            sc = None
        finally:
            return sc

    def make_supercells_with_defects(self, scaling_matrix, element):
        """
        Returns sequence of supercells in pymatgen.core.structure.Structure
        format, with each supercell containing an interstitial.
        First supercell has no defects.
        """
        sc_list_with_interstitial = []
        sc = self._structure.copy()
        sc.make_supercell(scaling_matrix)
        sc_list_with_interstitial.append(sc)
        for defect_site in self.enumerate_defectsites():
            sc_with_inter = self._supercell_with_defect(
                scaling_matrix, defect_site, element
            )
            if sc_with_inter:
                sc_list_with_interstitial.append(sc_with_inter)
        return sc_list_with_interstitial


class InterstitialAnalyzer(object):
    """
    Use GULP to compute the interstitial formation energy, relaxed structures.
    Works only for metal oxides due to the use of Buckingham Potentials.

    Args:
        inter: pymatgen.defects.point_defects.Interstitial
        el: Element name in short hand notation ("El")
        oxi_state: Oxidtation state
        scd: Super cell dimension as number. The scaling is equal along xyz.
    """
    def __init__(self, inter, el, oxi_state, scd=2):
        self._inter = inter
        self._el = el
        self._oxi_state = oxi_state
        self._scd = scd
        self._relax_energies = []
        self._norelax_energies = []
        self._relax_struct = []

    def get_energy(self, n, relax=True):
        """
        Formation Energy for nth symmetrically distinct interstitial.
        """
        if relax and not self._relax_energies:
            self._relax_analysis()
        if not relax and not self._norelax_energies:
            no_inter = self._inter.defectsite_count()
            inter_gulp_kw = ('qok',)
            val_dict = self._inter.struct_valences
            val_dict[self._el] = self._oxi_state  # If element not in structure
            scd = self._scd
            scale_mat = [[scd, 0, 0], [0, scd, 0], [0, 0, scd]]
            sc = self._inter.make_supercells_with_defects(scale_mat, self._el)
            blk_energy = get_energy_buckingham(sc[0])
            for i in range(1, no_inter + 1):
                inter_energy = get_energy_buckingham(
                    sc[i], keywords=inter_gulp_kw, valence_dict=val_dict
                )
                form_energy = inter_energy - blk_energy
                self._norelax_energies.append(form_energy)

        if relax:
            return self._relax_energies[n]
        else:
            return self._norelax_energies[n]

    def _relax_analysis(self):
        """
        Optimize interstitial structures
        """

        no_inter = self._inter.defectsite_count()
        inter_gulp_kw = ('optimise', 'conp', 'qok')
        val_dict = self._inter.struct_valences
        scd = self._scd
        scale_mat = [[scd, 0, 0], [0, scd, 0], [0, 0, scd]]
        sc = self._inter.make_supercells_with_defects(scale_mat, self._el)
        blk_energy, rlx_struct = get_energy_relax_structure_buckingham(sc[0])
        self._relax_struct.append(rlx_struct)
        val_dict[self._el] = self._oxi_state  # If element not in structure
        for i in range(1, no_inter + 1):
            energy, rlx_struct = get_energy_relax_structure_buckingham(
                sc[i], keywords=inter_gulp_kw, valence_dict=val_dict
            )
            form_energy = energy - blk_energy
            self._relax_energies.append(form_energy)
            self._relax_struct.append(rlx_struct)

    def get_relaxed_structure(self, n):
        """
        Optimized interstitial structure

        Args:
            n: Symmetrically distinct interstitial index

        .. note::

            To get relaxed bulk structure pass -1.
            -ve index will not work as expected.
        """
        if not self._relax_struct:
            self._relax_analysis()
        return self._relax_struct[n + 1]

    def get_percentage_volume_change(self, n):
        """
        Volume change after the introduction of interstitial

        Args:
            n: Symmetrically distinct interstitial index
        """
        if not self._relax_struct:
            self._relax_analysis()
        blk_struct = self._relax_struct[0]
        def_struct = self._relax_struct[n + 1:n + 2][0]
        del def_struct.sites[-1]
        rv = RelaxationAnalyzer(blk_struct, def_struct)
        return rv.get_percentage_volume_change()

    def get_percentage_lattice_parameter_change(self, n):
        """
        Lattice parameter change after the introduction of interstitial

        Args:
            n: Symmetrically distinct interstitial index
        """
        if not self._relax_struct:
            self._relax_analysis()
        blk_struct = self._relax_struct[0]
        def_struct = self._relax_struct[n + 1:n + 2][0]
        del def_struct.sites[-1]
        rv = RelaxationAnalyzer(blk_struct, def_struct)
        return rv.get_percentage_lattice_parameter_changes()

    def get_percentage_bond_distance_change(self, n):
        """
        Bond distance change after the introduction of interstitial

        Args:
            n: Symmetrically distinct interstitial index
        """
        if not self._relax_struct:
            self._relax_analysis()
        blk_struct = self._relax_struct[0]
        def_struct = self._relax_struct[n + 1:n + 2][0]
        del def_struct.sites[-1]
        #print def_struct
        rv = RelaxationAnalyzer(blk_struct, def_struct)
        return rv.get_percentage_bond_dist_changes()

    def relaxed_structure_match(self, i, j):
        """
        Check if the relaxed structures of two interstitials match

        Args:
            i: Symmetrically distinct interstitial index
            j: Symmetrically distinct interstitial index

        .. note::

            To use relaxed bulk structure pass -1.
            -ve index will not work as expected
        """
        if not self._relax_struct:
            self._relax_analysis()
        sm = StructureMatcher()
        struct1 = self._relax_struct[i + 1]
        struct2 = self._relax_struct[j + 1]
        return sm.fit(struct1, struct2)


class StructureRelaxer(object):
    def __init__(self, structure):
        self._unrelax_struct = structure
        self.relax()

    def relax(self):
        energy, rlx_struct = get_energy_relax_structure_buckingham(
            self._unrelax_struct)
        self._relax_struct = rlx_struct

    def get_relaxed_structure(self):
        return self._relax_struct


class InterstitialStructureRelaxer(object):
    """
    Performs structural relaxation for each interstitial supercell.

    Args:
        interstitial: Unrelaxed interstitial
        el: Species string in short notation
        oxi_state: Oxidation state of the element
        supercell_dim: Dimensions of super cell
    """

    def __init__(self, interstitial, el, oxi_state, supercell_dim=2):
        self._inter = interstitial
        self._scd = supercell_dim
        self._el = el
        self._oxi_state = oxi_state
        self._relax_structs = []
        self._relax_energies = []

    def relax(self):
        """
        Optimize interstitial structures
        """

        no_inter = self._inter.defectsite_count()
        inter_gulp_kw = ('optimise', 'conp', 'qok')
        val_dict = self._inter.struct_valences
        scd = self._scd
        scale_mat = [[scd, 0, 0], [0, scd, 0], [0, 0, scd]]
        sc = self._inter.make_supercells_with_defects(scale_mat, self._el)
        blk_energy, rlx_struct = get_energy_relax_structure_buckingham(sc[0])
        self._relax_structs.append(rlx_struct)
        self._relax_energies.append(blk_energy)
        val_dict[self._el] = self._oxi_state  # If element not in structure
        for i in range(1, no_inter + 1):
            try:
                energy, rlx_struct = get_energy_relax_structure_buckingham(
                    sc[i], keywords=inter_gulp_kw, valence_dict=val_dict
                )
                self._relax_energies.append(energy)
                self._relax_structs.append(rlx_struct)
            except:
                self._relax_energies.append(None)
                self._relax_structs.append(None)

        def is_empty(lst):
            for value in lst:
                if value:
                    return False
            return True

        if is_empty(self._relax_energies):
            raise IOError('Relaxation failed')

    def relaxed_structure_match(self, i, j):
        """
        Check if the relaxed structures of two interstitials match

        Args:
            i: Symmetrically distinct interstitial index
            j: Symmetrically distinct interstitial index

        .. note::

            Index 0 corresponds to bulk.
        """
        if not self._relax_structs:
            self.relax()
        sm = StructureMatcher()
        struct1 = self._relax_structs[i]
        struct2 = self._relax_structs[j]
        return sm.fit(struct1, struct2)

    def relaxed_energy_match(self, i, j):
        """
        Check if the relaxed energies of two interstitials match

        Args:
            i: Symmetrically distinct interstitial index
            j: Symmetrically distinct interstitial index

        .. note::

            Index 0 corresponds to bulk.
        """
        if not self._relax_energies:
            self.relax()
        energy1 = self._relax_energies[i]
        energy2 = self._relax_energies[j]
        return energy1 == energy2

    def get_relaxed_structure(self, n):
        """
        Get the relaxed structure of nth symmetrically distinct interstitial.

        Args:
            n: Symmetrically distinct interstitial index

        .. note::

            0 corresponds to relaxed bulk structure
        """
        if not self._relax_structs:
            self.relax()
        return self._relax_structs[n]

    def get_relaxed_energy(self, n):
        """
        Get the relaxed structure of nth symmetrically distinct interstitial.

        Args:
            n: Symmetrically distinct interstitial index

        .. note::

            0 corresponds to relaxed bulk structure
        """
        if not self._relax_energies:
            self.relax()
        return self._relax_energies[n]

    def get_relaxed_interstitial(self):
        """
        Get the relaxed structure of nth symmetrically distinct interstitial.

        Args:
            n: Symmetrically distinct interstitial index
        """
        if not self._relax_energies:
            self.relax()
        energies = self._relax_energies[:]
        structs = self._relax_structs[:]
        distinct_energy_set = set(energies[1:])  # only interstitial energies
        if None in distinct_energy_set:
            distinct_energy_set.remove(None)
        distinct_structs = [structs[0]]         # bulk
        distinct_energies = [energies[0]]
        for energy in distinct_energy_set:
            ind = energies.index(energy)
            distinct_structs.append(structs[ind])
            distinct_energies.append(energies[ind])
        return RelaxedInterstitial(
            distinct_structs, distinct_energies, self._inter.struct_valences
        )


class RelaxedInterstitial(object):
    """
    Stores the relaxed supercell structures for each interstitial
    Used to compute formation energies, displacement of atoms near the
    the interstitial.

    Args:
        struct_list: List of structures(supercells). The first structure should
            represent relaxed bulk structure and the subsequent ones
            interstitial structures (with the extra interstitial site
            appended at the end).
        energy_list: List of energies for relaxed interstitial structures.
            The first energy should correspond to bulk structure
        valence_dict: Valences of elements in dictionary form
    """

    def __init__(self, struct_list, energy_list, valence_dict):
        self._blk_struct = struct_list[0]
        struct_list.pop(0)
        self._structs = struct_list
        self._blk_energy = energy_list[0]
        energy_list.pop(0)
        self._energies = energy_list
        self._valence_dict = valence_dict

        self._coord_no = []
        self._coord_sites = []
        self._coord_charge_no = []

    def formation_energy(self, n, chem_pot=0):
        """
        Compute the interstitial formation energy

        Args:
            n: Index of interstitials
            chem_pot: Chemical potential of interstitial site element.
                If not given, assumed as zero. The user is strongly
                urged to supply the chemical potential value
        """
        return self._energies[n] - self._blk_energy - chem_pot

    def get_percentage_volume_change(self, n):
        """
        Volume change after the introduction of interstitial

        Args:
            n: index of interstitials
        """
        def_struct = self._structs[n:n + 1][0]
        del def_struct.sites[-1]
        rv = RelaxationAnalyzer(self._blk_struct, def_struct)
        return rv.get_percentage_volume_change()

    def get_percentage_lattice_parameter_change(self, n):
        """
        Lattice parameter change after the introduction of interstitial

        Args:
            n: index of interstitials
        """
        def_struct = self._structs[n:n + 1][0]  # copy
        del def_struct.sites[-1]
        rv = RelaxationAnalyzer(self._blk_struct, def_struct)
        return rv.get_percentage_lattice_parameter_changes()

    def get_percentage_bond_distance_change(self, n):
        """
        Bond distance change after the introduction of interstitial.

        Args:
            n: index of interstitials
        """
        def_struct = self._structs[n:n + 1][0]  # copy
        del def_struct.sites[-1]
        rv = RelaxationAnalyzer(self._blk_struct, def_struct)
        return rv.get_percentage_bond_dist_changes()

    def get_bulk_structure(self):
        """
        Return relaxed bulk structure
        """
        return self._blk_struct

    def get_interstitial_structure(self, n):
        """
        Return relaxed bulk structure
        """
        return self._structs[n]

    def defect_count(self):
        """
        Returns the number of distinct interstitials
        """
        return len(self._structs)

    def get_defectsite(self, n):
        """
        Returns the defect site of nth interstitial.

        Args:
            n: Index of interstitial
        """
        return self._structs[n][-1]

    def get_coordination_number(self, n):
        """
        Coordination number for nth interstitial.

        Args:
            n: Index of interstitials
        """
        if not self._coord_no:
            self._coord_find()
        return self._coord_no[n]

    def get_charge_coordination_number(self, n):
        """
        Charge coordination number for nth interstitial.

        Args:
            n: Index of interstitials
        """
        if not self._coord_charge_no:
            self._coord_find()
        return self._coord_charge_no[n]

    def get_coordinated_sites(self, n):
        """
        Coordinated sites for nth interstitial.

        Args:
            n: Index of interstitials
        """
        if not self._coord_sites:
            self._coord_find()
        return self._coord_sites[n]

    def get_coordinated_bulk_sites(self, n):
        """
        Bulk sites corresponding to the coordinated sites for nth interstitial.

        Args:
            n: Index of interstitials
        """
        blk_sites = []
        for site in self.get_coordinated_sites(n):
            site_index = self._structs[n].sites.index(site)
            blk_sites.append(self._blk_struct[site_index])
        return blk_sites

    def get_coordinated_site_displacement(self, n):
        """
        Compute the total displacement of coordinated sites from the
        interstitial sites during the relaxation

        Args:
            n: Index  of defect site
        """
        coord_sites = self.get_coordinated_sites(n)
        coord_blk_sites = self.get_coordinated_bulk_sites(n)
        dist_sum = 0
        for i in range(len(coord_sites)):
            dist_sum += coord_sites[i].distance_from_point(coord_blk_sites[i])
            # How to compute the average?
        return dist_sum


    def _coord_find(self):
        """
        calls VoronoiCoordFinder to compute the coordination number,
        coordination charge
        """
        for i in range(self.defect_count()):
            struct = self._structs[i].copy()
            coord_finder = VoronoiCoordFinder(struct)
            self._coord_no.append(coord_finder.get_coordination_number(-1))
            self._coord_sites.append(coord_finder.get_coordinated_sites(-1))
            coord_chrg = 0
            for site, weight in coord_finder.get_voronoi_polyhedra(-1).items():
                coord_chrg += weight * self._valence_dict[site.species_string]
            self._coord_charge_no.append(coord_chrg)


def symmetry_reduced_voronoi_nodes(
        structure, rad_dict, high_accuracy_flag=False, symm_flag=True):
    """
    Obtain symmetry reduced voronoi nodes using Zeo++ and
    pymatgen.symmetry.finder.SpacegroupAnalyzer

    Args:
        strucutre: pymatgen Structure object
        rad_dict: Dictionary containing radii of spcies in the structure
        high_accuracy_flag: Flag denotting whether to use high accuracy version of Zeo++
        symm_flag: Flag denoting whether to return symmetrically distinct sites only

    Returns:
        Symmetrically distinct voronoi nodes as pymatgen Strucutre
    """

    def add_closest_equiv_site(dist_sites, equiv_sites):
        if not dist_sites:
            dist_sites.append(equiv_sites[0])
        else:
            avg_dists = []
            for site in equiv_sites:
                dists = [site.distance(dst_site, jimage=[0, 0, 0])
                         for dst_site in dist_sites]
                avg_dist = sum(dists) / len(dist_sites)
                avg_dists.append(avg_dist)

            min_avg_dist = min(avg_dists)
            ind = avg_dists.index(min_avg_dist)
            dist_sites.append(equiv_sites[ind])

    def cmp_memoize_last_site(f): #Compares and stores last site
        def not_duplicates(site1, site2):
            if site1.distance(site2) < 1e-5:
                return False
            else:
                return True

        cmp_memoize_last_site.cache = None
        def helper(x):
            if not cmp_memoize_last_site.cache: 
                cmp_memoize_last_site.cache = f(x)
                return True
            y = f(x)
            if not_duplicates(cmp_memoize_last_site.cache, y):
                cmp_memoize_last_site.cache = y
                return True
            else:
                return False
        return helper

    @cmp_memoize_last_site
    def check_not_duplicates(site):
        return site


    if not symm_flag:
        if not high_accuracy_flag:
            vor_node_struct, vor_edgecenter_struct, vor_facecenter_struct = \
                get_voronoi_nodes(structure, rad_dict)
            return vor_node_struct.sites, vor_edgecenter_struct.sites, \
                   vor_facecenter_struct.sites
        else:
            # Only the nodes are from high accuracy voronoi decomposition
            vor_node_struct = \
                    get_high_accuracy_voronoi_nodes(structure, rad_dict)
            # Before getting the symmetry, remove the duplicates
            vor_node_struct.sites.sort(key = lambda site: site.voronoi_radius)
            #print type(vor_node_struct.sites[0])
            dist_sites = filter(check_not_duplicates, vor_node_struct.sites)
            return dist_sites, None, None

    if not high_accuracy_flag:
        vor_node_struct, vor_edgecenter_struct, vor_facecenter_struct = \
            get_voronoi_nodes(structure, rad_dict)
        vor_node_symmetry_finder = SpacegroupAnalyzer(vor_node_struct, symprec=1e-1)
        vor_node_symm_struct = vor_node_symmetry_finder.get_symmetrized_structure()
        node_equiv_sites_list = vor_node_symm_struct.equivalent_sites

        node_dist_sites = []
        for equiv_sites in node_equiv_sites_list:
            add_closest_equiv_site(node_dist_sites, equiv_sites)

        vor_edge_symmetry_finder = SpacegroupAnalyzer(
            vor_edgecenter_struct, symprec=1e-1)
        vor_edge_symm_struct = vor_edge_symmetry_finder.get_symmetrized_structure()
        edgecenter_equiv_sites_list = vor_edge_symm_struct.equivalent_sites

        edgecenter_dist_sites = []
        for equiv_sites in edgecenter_equiv_sites_list:
            add_closest_equiv_site(edgecenter_dist_sites, equiv_sites)
        if not edgecenter_equiv_sites_list:     # Fix this so doesn't arise
            edgecenter_dist_sites = vor_edgecenter_struct.sites

        vor_fc_symmetry_finder = SpacegroupAnalyzer(
                        vor_facecenter_struct, symprec=1e-1)
        vor_fc_symm_struct = vor_fc_symmetry_finder.get_symmetrized_structure()
        facecenter_equiv_sites_list = vor_fc_symm_struct.equivalent_sites

        facecenter_dist_sites = []
        for equiv_sites in facecenter_equiv_sites_list:
            add_closest_equiv_site(facecenter_dist_sites, equiv_sites)
        if not facecenter_equiv_sites_list:     # Fix this so doesn't arise
            facecenter_dist_sites = vor_facecenter_struct.sites

        return node_dist_sites, edgecenter_dist_sites, facecenter_dist_sites
    else:
        # Only the nodes are from high accuracy voronoi decomposition
        vor_node_struct = \
                get_high_accuracy_voronoi_nodes(structure, rad_dict)

        # Before getting the symmetry, remove the duplicates
        vor_node_struct.sites.sort(key = lambda site: site.voronoi_radius)
        #print type(vor_node_struct.sites[0])
        dist_sites = list(filter(check_not_duplicates, vor_node_struct.sites))

        # Ignore symmetry from ha voronoi nodes
        # Increase the symmetry precision to 0.25
        #spg = SpacegroupAnalyzer(structure,symprec=1e-1).get_spacegroup()
        
        # Remove symmetrically equivalent sites
        #i = 0
        #while (i < len(dist_sites)-1):
        #    sites1 = [dist_sites[i]]
        #    sites2 = [dist_sites[i+1]]
        #    if spg.are_symmetrically_equivalent(sites1,sites2):
        #        del dist_sites[i+1]
        #    else:
        #        i = i+1

        node_dist_sites = dist_sites
        return (node_dist_sites, None, None)

        #vor_edge_symmetry_finder = SpacegroupAnalyzer(
        #    vor_edgecenter_struct, symprec=1e-1)
        #vor_edge_symm_struct = vor_edge_symmetry_finder.get_symmetrized_structure()
        #edgecenter_equiv_sites_list = vor_edge_symm_struct.equivalent_sites

        #edgecenter_dist_sites = []
        #for equiv_sites in edgecenter_equiv_sites_list:
        #    add_closest_equiv_site(edgecenter_dist_sites, equiv_sites)
        #if not edgecenter_equiv_sites_list:     
        #    edgecenter_dist_sites = vor_edgecenter_struct.sites

        #vor_fc_symmetry_finder = SpacegroupAnalyzer(
        #                vor_facecenter_struct, symprec=1e-1)
        #vor_fc_symm_struct = vor_fc_symmetry_finder.get_symmetrized_structure()
        #facecenter_equiv_sites_list = vor_fc_symm_struct.equivalent_sites

        #facecenter_dist_sites = []
        #for equiv_sites in facecenter_equiv_sites_list:
        #    add_closest_equiv_site(facecenter_dist_sites, equiv_sites)
        #if not facecenter_equiv_sites_list:     
        #    facecenter_dist_sites = vor_facecenter_struct.sites

        #return node_dist_sites, edgecenter_dist_sites, facecenter_dist_sites



# coding: utf-8
# Copyright (c) Pymatgen Development Team.
# Distributed under the terms of the MIT License.

from __future__ import division, unicode_literals

"""
This module contains the main object used to identify the coordination environments in a given structure.
"""

__author__ = "David Waroquiers"
__copyright__ = "Copyright 2012, The Materials Project"
__credits__ = "Geoffroy Hautier"
__version__ = "2.0"
__maintainer__ = "David Waroquiers"
__email__ = "david.waroquiers@gmail.com"
__date__ = "Feb 20, 2016"

import itertools
import logging
import time

from numpy.linalg import svd
from numpy.linalg import norm
from numpy import transpose
from pymatgen.core.structure import Structure
from pymatgen.core.lattice import Lattice
from pymatgen.core.periodic_table import Specie
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from pymatgen.analysis.bond_valence import BVAnalyzer
import numpy as np

from random import shuffle

from pymatgen.analysis.chemenv.utils.coordination_geometry_utils import \
    vectorsToMatrix, rotateCoords, Plane
from pymatgen.analysis.chemenv.utils.coordination_geometry_utils import \
    matrixMultiplication
from pymatgen.analysis.chemenv.utils.coordination_geometry_utils import \
    collinear, separation_in_list
from pymatgen.analysis.chemenv.utils.coordination_geometry_utils import \
    sort_separation
from pymatgen.analysis.chemenv.coordination_environments.coordination_geometries import \
    AllCoordinationGeometries
from pymatgen.analysis.chemenv.coordination_environments.coordination_geometries import \
    EXPLICIT_PERMUTATIONS
from pymatgen.analysis.chemenv.coordination_environments.coordination_geometries import \
    SEPARATION_PLANE
from pymatgen.analysis.chemenv.coordination_environments.structure_environments import \
    ChemicalEnvironments
from pymatgen.analysis.chemenv.coordination_environments.structure_environments import \
    StructureEnvironments
from pymatgen.analysis.chemenv.coordination_environments.voronoi import \
    DetailedVoronoiContainer

debug = False
DIST_TOLERANCES = [0.02, 0.05, 0.1, 0.2, 0.3]


class AbstractGeometry(object):
    """
    Class used to describe a geometry (perfect or distorted)
    """

    def __init__(self, central_site=None, bare_coords=None,
                 centering_type='standard',
                 include_central_site_in_centroid=False):
        """
        Constructor for the abstract geometry
        :param central_site: Coordinates of the central site
        :param bare_coords: Coordinates of the neighbors of the central site
        :param centering_type: How to center the abstract geometry
        :param include_central_site_in_centroid: When the centering is on the centroid, the central site is included
         if this parameter is set to True.
        :raise: ValueError if the parameters are not consistent
        """
        self.bare_centre = np.array(central_site)
        self.bare_points_without_centre = [np.array(bc) for bc in bare_coords]
        self.bare_points_with_centre = [np.array(central_site)]
        self.bare_points_with_centre.extend(
            [np.array(bc) for bc in bare_coords])

        self.centroid_without_centre = np.mean(self.bare_points_without_centre,
                                               axis=0)
        self.centroid_with_centre = np.mean(self.bare_points_with_centre,
                                            axis=0)

        self._points_wcs_csc = [pp - self.bare_centre for pp in
                                self.bare_points_with_centre]
        self._points_wocs_csc = [pp - self.bare_centre for pp in
                                 self.bare_points_without_centre]
        self._points_wcs_ctwcc = [pp - self.centroid_with_centre for pp in
                                  self.bare_points_with_centre]
        self._points_wocs_ctwcc = [pp - self.centroid_with_centre for pp in
                                   self.bare_points_without_centre]
        self._points_wcs_ctwocc = [pp - self.centroid_without_centre for pp in
                                   self.bare_points_with_centre]
        self._points_wocs_ctwocc = [pp - self.centroid_without_centre for pp in
                                    self.bare_points_without_centre]

        self.centering_type = centering_type
        self.include_central_site_in_centroid = include_central_site_in_centroid
        self.bare_central_site = central_site
        if centering_type == 'standard':
            if len(bare_coords) < 5:
                if include_central_site_in_centroid:
                    raise ValueError(
                        "The center is the central site, no calculation of the centroid,"
                        "variable include_central_site_in_centroid should be set to False")
                if central_site is None:
                    raise ValueError(
                        "Centering_type is central_site, the central site should be given")
                self.centre = np.array(central_site)
            else:
                total = np.zeros(3, np.float)
                for pp in bare_coords:
                    total += pp
                if include_central_site_in_centroid:
                    if central_site is None:
                        raise ValueError(
                            "The centroid includes the central site but no central site is given")
                    total += central_site
                    self.centre = total / (np.float(len(bare_coords)) + 1.0)
                else:
                    self.centre = total / np.float(len(bare_coords))
        elif centering_type == 'central_site':
            if include_central_site_in_centroid:
                raise ValueError(
                    "The center is the central site, no calculation of the centroid,"
                    "variable include_central_site_in_centroid should be set to False")
            if central_site is None:
                raise ValueError(
                    "Centering_type is central_site, the central site should be given")
            self.centre = np.array(central_site)
        elif centering_type == 'centroid':
            total = np.zeros(3, np.float)
            for pp in bare_coords:
                total += pp
            if include_central_site_in_centroid:
                if central_site is None:
                    raise ValueError(
                        "The centroid includes the central site but no central site is given")
                total += central_site
                self.centre = total / (np.float(len(bare_coords)) + 1.0)
            else:
                self.centre = total / np.float(len(bare_coords))
        self._bare_coords = bare_coords
        self._coords = list()
        for ip, pp in enumerate(bare_coords):
            self._coords.append(np.array(pp) - self.centre)
        self.central_site = self.bare_central_site - self.centre
        self.coords = self._coords
        self.bare_coords = self._bare_coords

    def __str__(self):
        """
        String representation of the AbstractGeometry
        :return: String representation of the AbstractGeometry
        """
        outs = [
            '\nAbstract Geometry with {n} points :'.format(n=len(self.coords))]
        for pp in self.coords:
            outs.append('  {pp}'.format(pp=pp))
        if self.centering_type == 'standard':
            if self.include_central_site_in_centroid:
                outs.append(
                    'Points are referenced to the central site for coordination numbers < 5'
                    ' and to the centroid (calculated with the central site) for coordination'
                    ' numbers >= 5 : {c}\n'.format(c=self.centre))
            else:
                outs.append(
                    'Points are referenced to the central site for coordination numbers < 5'
                    ' and to the centroid (calculated without the central site) for coordination'
                    ' numbers >= 5 : {c}\n'.format(c=self.centre))
        elif self.centering_type == 'central_site':
            outs.append(
                'Points are referenced to the central site : {c}\n'.format(
                    c=self.centre))
        elif self.centering_type == 'centroid':
            if self.include_central_site_in_centroid:
                outs.append('Points are referenced to the centroid'
                            ' (calculated with the central site) :\n  {c}\n'.format(
                    c=self.centre))
            else:
                outs.append('Points are referenced to the centroid'
                            ' (calculated without the central site) :\n  {c}\n'.format(
                    c=self.centre))
        return '\n'.join(outs)

    @classmethod
    def from_cg(cls, cg, centering_type='standard',
                include_central_site_in_centroid=False):
        central_site = cg.get_central_site()
        bare_coords = [np.array(pt, np.float) for pt in cg.points]
        return cls(central_site=central_site, bare_coords=bare_coords,
                   centering_type=centering_type,
                   include_central_site_in_centroid=include_central_site_in_centroid)

    def points_wcs_csc(self, permutation=None):
        if permutation is None:
            return self._points_wcs_csc
        perm = [0]
        perm.extend([pp + 1 for pp in permutation])
        return [self._points_wcs_csc[ii] for ii in perm]

    def points_wocs_csc(self, permutation=None):
        if permutation is None:
            return self._points_wocs_csc
        return [self._points_wocs_csc[ii] for ii in permutation]

    def points_wcs_ctwcc(self, permutation=None):
        if permutation is None:
            return self._points_wcs_ctwcc
        perm = [0]
        perm.extend([pp + 1 for pp in permutation])
        return [self._points_wcs_ctwcc[ii] for ii in perm]

    def points_wocs_ctwcc(self, permutation=None):
        if permutation is None:
            return self._points_wocs_ctwcc
        return [self._points_wocs_ctwcc[ii] for ii in permutation]

    def points_wcs_ctwocc(self, permutation=None):
        if permutation is None:
            return self._points_wcs_ctwocc
        perm = [0]
        perm.extend([pp + 1 for pp in permutation])
        return [self._points_wcs_ctwocc[ii] for ii in perm]

    def points_wocs_ctwocc(self, permutation=None):
        if permutation is None:
            return self._points_wocs_ctwocc
        return [self._points_wocs_ctwocc[ii] for ii in permutation]


def symmetry_measure(points_distorted, points_perfect):
    """
    Computes the continuous symmetry measure of the (distorted) set of points "points_distorted" with respect to the
    (perfect) set of points "points_perfect".
    :param points_distorted: List of points describing a given (distorted) polyhedron for which the symmetry measure
                             has to be computed with respect to the model polyhedron described by the list of points
                             "points_perfect".
    :param points_perfect: List of "perfect" points describing a given model polyhedron.
    :return: The continuous symmetry measure of the distorted polyhedron with respect to the perfect polyhedron
    """
    # When there is only one point, the symmetry measure is 0.0 by definition
    if len(points_distorted) == 1:
        return 0.0
    # Find the rotation matrix that aligns the distorted points to the perfect points in a least-square sense.
    rot = find_rotation(points_distorted=points_distorted,
                        points_perfect=points_perfect)
    # Find the scaling factor between the distorted points and the perfect points in a least-square sense.
    scaling_factor, rotated_coords = find_scaling_factor(
        points_distorted=points_distorted,
        points_perfect=points_perfect,
        rot=rot)
    # Compute the continuous symmetry measure [see Eq. 1 in Pinsky et al., Inorganic Chemistry 37, 5575 (1998)]
    num = 0.0
    denom = 0.0
    for ip, pp in enumerate(points_perfect):
        rotated_coords[ip] = scaling_factor * rotated_coords[ip]
        diff = (pp - rotated_coords[ip])
        num += np.sum(diff * diff)
        denom += np.sum(pp * pp)
    return num / denom * 100.0


def find_rotation(points_distorted, points_perfect):
    """
    This finds the rotation matrix that aligns the (distorted) set of points "points_distorted" with respect to the
    (perfect) set of points "points_perfect" in a least-square sense.
    :param points_distorted: List of points describing a given (distorted) polyhedron for which the rotation that
                             aligns these points in a least-square sense to the set of perfect points "points_perfect"
    :param points_perfect: List of "perfect" points describing a given model polyhedron.
    :return: The rotation matrix
    """
    isexact = True
    for ip, pp in enumerate(points_distorted):
        if not np.allclose(pp, points_perfect[ip]):
            isexact = False
            break
    if isexact:
        rot = np.eye(3)
        return rot
    H = np.zeros([3, 3], np.float)
    for ip, pp in enumerate(points_distorted):
        H += vectorsToMatrix(pp, points_perfect[ip])
    [U, S, Vt] = svd(H)
    rot = matrixMultiplication(transpose(Vt), transpose(U))
    return rot


def find_scaling_factor(points_distorted, points_perfect, rot):
    """
    This finds the scaling factor between the (distorted) set of points "points_distorted" and the
    (perfect) set of points "points_perfect" in a least-square sense.
    :param points_distorted: List of points describing a given (distorted) polyhedron for which the scaling factor has
                             to be obtained.
    :param points_perfect: List of "perfect" points describing a given model polyhedron.
    :param rot: The rotation matrix
    :return: The scaling factor between the two structures and the rotated set of (distorted) points.
    """
    rotated_coords = rotateCoords(points_distorted, rot)
    num = np.sum([np.dot(rc, points_perfect[ii]) for ii, rc in
                  enumerate(rotated_coords)])
    denom = np.sum([np.dot(rc, rc) for rc in rotated_coords])
    return num / denom, rotated_coords


class LocalGeometryFinder(object):
    """
    Main class used to find the local environments in a structure
    """
    DEFAULT_BVA_DISTANCE_SCALE_FACTOR = 1.0
    BVA_DISTANCE_SCALE_FACTORS = {'experimental': 1.0, 'GGA_relaxed': 1.015,
                                  'LDA_relaxed': 0.995}
    DEFAULT_SPG_ANALYZER_OPTIONS = {'symprec': 1e-3, 'angle_tolerance': 5}
    STRUCTURE_REFINEMENT_NONE = 'none'
    STRUCTURE_REFINEMENT_REFINED = 'refined'
    STRUCTURE_REFINEMENT_SYMMETRIZED = 'symmetrized'

    def __init__(self, permutations_safe_override=False,
                 plane_ordering_override=True, debug_level=None,
                 plane_safe_permutations=False,
                 logfile='chemenv_local_geometry_finder.log',
                 only_symbols=None):
        """
        Constructor for the LocalGeometryFinder, initializes the list of coordination geometries
        :param permutations_safe_override: If set to True, all permutations are tested (very time-consuming for large
        coordination numbers!)
        :param plane_ordering_override: If set to False, the ordering of the points in the plane is disabled
        """
        self.cg = AllCoordinationGeometries(
            permutations_safe_override=permutations_safe_override,
            only_symbols=only_symbols)
        self.permutations_safe_override = permutations_safe_override
        self.plane_ordering_override = plane_ordering_override
        self.plane_safe_permutations = plane_safe_permutations
        self.setup_parameters(centering_type='centroid',
                              include_central_site_in_centroid=True,
                              bva_distance_scale_factor=None,
                              structure_refinement='refined')

    def setup_parameters(self, centering_type='standard',
                         include_central_site_in_centroid=False,
                         bva_distance_scale_factor=None,
                         structure_refinement=STRUCTURE_REFINEMENT_REFINED,
                         spg_analyzer_options=None):
        """
        Setup of the parameters for the coordination geometry finder. A reference point for the geometries has to be
        chosen. This can be the centroid of the structure (including or excluding the atom for which the coordination
        geometry is looked for) or the atom itself. In the 'standard' centering_type, the reference point is the central
        atom for coordination numbers 1, 2, 3 and 4 and the centroid for coordination numbers > 4.
        :param centering_type: Type of the reference point (centering) 'standard', 'centroid' or 'central_site'
        :param include_central_site_in_centroid: In case centering_type is 'centroid', the central site is included if
                                                 this value is set to True.
        :param bva_distance_scale_factor: Scaling factor for the bond valence analyzer (this might be different whether
                                          the structure is an experimental one, an LDA or a GGA relaxed one, or any
                                          other relaxation scheme (where under- or over-estimation of bond lengths
                                          is known).
        :param structure_refinement: Refinement of the structure. Can be "none", "refined" or "symmetrized".
        :param spg_analyzer_options: Options for the SpaceGroupAnalyzer (dictionary specifying "symprec"
                                     and "angle_tolerance". See pymatgen's SpaceGroupAnalyzer for more information.
        """
        self.centering_type = centering_type
        self.include_central_site_in_centroid = include_central_site_in_centroid
        if bva_distance_scale_factor is not None:
            self.bva_distance_scale_factor = bva_distance_scale_factor
        else:
            self.bva_distance_scale_factor = self.DEFAULT_BVA_DISTANCE_SCALE_FACTOR
        self.structure_refinement = structure_refinement
        if spg_analyzer_options is None:
            self.spg_analyzer_options = self.DEFAULT_SPG_ANALYZER_OPTIONS
        else:
            self.spg_analyzer_options = spg_analyzer_options

    def setup_parameter(self, parameter, value):
        """
        Setup of one specific parameter to the given value. The other parameters are unchanged. See setup_parameters
        method for the list of possible parameters
        :param parameter: Parameter to setup/update
        :param value: Value of the parameter
        """
        self.__dict__[parameter] = value

    def setup_structure(self, structure):
        """
        Sets up the structure for which the coordination geometries have to be identified. The structure is analyzed
        with the space group analyzer and a refined structure is used
        :param structure: A pymatgen Structure
        :param
        """
        self.initial_structure = structure.copy()
        if self.structure_refinement == self.STRUCTURE_REFINEMENT_NONE:
            self.structure = structure.copy()
            self.spg_analyzer = None
            self.symmetrized_structure = None
        else:
            self.spg_analyzer = SpacegroupAnalyzer(self.initial_structure,
                                                   symprec=
                                                   self.spg_analyzer_options[
                                                       'symprec'],
                                                   angle_tolerance=
                                                   self.spg_analyzer_options[
                                                       'angle_tolerance'])
            if self.structure_refinement == self.STRUCTURE_REFINEMENT_REFINED:
                self.structure = self.spg_analyzer.get_refined_structure()
                self.symmetrized_structure = None
            elif self.structure_refinement == self.STRUCTURE_REFINEMENT_SYMMETRIZED:
                self.structure = self.spg_analyzer.get_refined_structure()
                self.spg_analyzer_refined = SpacegroupAnalyzer(self.structure,
                                                               symprec=
                                                               self.spg_analyzer_options[
                                                                   'symprec'],
                                                               angle_tolerance=
                                                               self.spg_analyzer_options
                                                               [
                                                                   'angle_tolerance'])
                self.symmetrized_structure = self.spg_analyzer_refined.get_symmetrized_structure()

    def get_structure(self):
        """
        Returns the pymatgen Structure that has been setup for the identification of geometries (the initial one
        might have been refined/symmetrized using the SpaceGroupAnalyzer).
        :return: The pymatgen Structure that has been setup for the identification of geometries (the initial one
        might have been refined/symmetrized using the SpaceGroupAnalyzer).
        """
        return self.structure

    def set_structure(self, lattice, species, coords, coords_are_cartesian):
        """
        Sets up the pymatgen structure for which the coordination geometries have to be identified starting from the
        lattice, the species and the coordinates
        :param lattice: The lattice of the structure
        :param species: The species on the sites
        :param coords: The coordinates of the sites
        :param coords_are_cartesian: If set to True, the coordinates are given in cartesian coordinates
        """
        self.setup_structure(
            Structure(lattice, species, coords, coords_are_cartesian))

    def compute_structure_environments_detailed_voronoi(self,
                                                        excluded_atoms=None,
                                                        only_atoms=None,
                                                        only_cations=True,
                                                        only_indices=None,
                                                        source_structure_valence_fallback=False,
                                                        no_valence_exclude_atoms_fallback=None,
                                                        maximum_distance_factor=None,
                                                        minimum_angle_factor=None,
                                                        max_cn=None):
        """
        Computes and returns the StructureEnvironments object containing all the information about the coordination
        environments in the structure
        :param excluded_atoms: Atoms for which the coordination geometries does not have to be identified
        :param only_atoms: If not set to None, atoms for which the coordination geometries have to be identified
        :return: The StructureEnvironments object containing all the information about the coordination
        environments in the structure
        """

        # Bond valence analysis to get approximated valences
        logging.info('Getting valences using BVAnalyzer')
        bva = BVAnalyzer(distance_scale_factor=self.bva_distance_scale_factor)
        self.info = {}
        try:
            self.bva_valences = bva.get_valences(self.structure)
            self.info['valences'] = 'bva'
        except:
            self.bva_valences = 'undefined'
            self.info['valences'] = 'undefined'
        self.valences = self.bva_valences

        # Get a list of indices of unequivalent sites from the initial structure
        if (
                self.structure_refinement == self.STRUCTURE_REFINEMENT_SYMMETRIZED and
                len(self.symmetrized_structure.equivalent_sites) > 0):
            logging.info('Symmetrizing and refining structure')
            indices = []
            ind_eqsites_found = []
            self.equivalent_sites = self.symmetrized_structure.equivalent_sites
            self.struct_sites_to_irreducible_site_list_map = [-1] * len(
                self.structure)
            self.sites_map = [-1] * len(self.structure)
            eqsite_found = [-1] * len(
                self.symmetrized_structure.equivalent_sites)
            for isite, site in enumerate(self.structure):
                for ieqsites, eqsites in enumerate(
                        self.symmetrized_structure.equivalent_sites):
                    if site in eqsites:
                        self.struct_sites_to_irreducible_site_list_map[
                            isite] = ieqsites
                        if ieqsites not in ind_eqsites_found:
                            indices.append(isite)
                            ind_eqsites_found.append(ieqsites)
                            myieqsite = isite
                            eqsite_found[ieqsites] = myieqsite
                        else:
                            myieqsite = eqsite_found[ieqsites]
                self.sites_map[isite] = myieqsite
        else:
            self.equivalent_sites = [[site] for site in self.structure]
            self.struct_sites_to_irreducible_site_list_map = list(
                range(len(self.structure)))
            self.sites_map = list(range(len(self.structure)))
            indices = list(range(len(self.structure)))

        if source_structure_valence_fallback and self.bva_valences == 'undefined':
            dummyspoccu = self.structure[0].species_and_occu
            ok = False
            if isinstance(dummyspoccu.keys()[0], Specie):
                self.valences = []
                for isite, site in enumerate(self.structure):
                    oxi_state = sum([frac * sp.oxi_state for sp, frac in
                                     site.species_and_occu.items()])
                    self.valences.append(oxi_state)
                if any([val != 0 for val in self.valences]):
                    ok = True
                else:
                    self.valences = 'undefined'
            if not ok:
                if no_valence_exclude_atoms_fallback is not None:
                    if excluded_atoms is None:
                        excluded_atoms = no_valence_exclude_atoms_fallback
                    else:
                        for atom in no_valence_exclude_atoms_fallback:
                            if atom not in excluded_atoms:
                                excluded_atoms.append(atom)

        # Get list of unequivalent sites with valence >= 0
        if only_cations and self.valences != 'undefined':
            sites_indices = [isite for isite in indices if
                             self.valences[isite] >= 0]
        else:
            sites_indices = [isite for isite in indices]

        # Include atoms that are in the list of "only_atoms" if it is provided
        if only_atoms is not None:
            sites_indices = [isite for isite in sites_indices
                             if any([at in [sp.symbol for sp in self.structure[
                    isite].species_and_occu]
                                     for at in only_atoms])]

        # Exclude atoms that are in the list of excluded atoms
        if excluded_atoms:
            sites_indices = [isite for isite in sites_indices
                             if not any([at in [sp.symbol for sp in
                                                self.structure[
                                                    isite].species_and_occu]
                                         for at in excluded_atoms])]

        if only_indices is not None:
            sites_indices = [isite for isite in indices if
                             isite in only_indices]

        # Get the VoronoiContainer for this list of unequivalent sites with valence >= 0
        logging.info('Getting DetailedVoronoiContainer')
        self.detailed_voronoi = DetailedVoronoiContainer(self.structure,
                                                         isites=sites_indices,
                                                         valences=self.valences,
                                                         maximum_distance_factor=maximum_distance_factor,
                                                         minimum_angle_factor=minimum_angle_factor)
        logging.info('DetailedVoronoiContainer has been set up')

        ce_list = []
        skipped = []
        logging.info('Computing structure environments')
        tse1 = time.clock()
        for isite in range(len(self.structure)):
            if isite not in sites_indices:
                logging.info(' ... in site #{:d} ({}) : skipped'.format(isite,
                                                                        self.structure[
                                                                            isite].species_string))
                skipped.append(isite)
                ce_list.append(None)
                continue
            logging.info(' ... in site #{:d} ({})'.format(isite, self.structure[
                isite].species_string))
            t1 = time.clock()
            coords = self.detailed_voronoi.unique_coordinations(isite)

            ce_dict = {}
            for cn in coords:
                if max_cn is not None and cn > max_cn:
                    continue
                ce_dict[cn] = []
                for i_nlist, nlist_tuple in enumerate(coords[cn]):
                    neighb_list = nlist_tuple[0]
                    ce = ChemicalEnvironments()
                    mycoords = [st.coords for st in neighb_list]
                    self.setup_local_geometry(isite, coords=mycoords)
                    cncgsm = self.get_coordination_symmetry_measures()
                    for cg in cncgsm:
                        other_csms = {
                            'csm_wocs_ctwocc': cncgsm[cg]['csm_wocs_ctwocc'],
                            'csm_wocs_ctwcc': cncgsm[cg]['csm_wocs_ctwcc'],
                            'csm_wocs_csc': cncgsm[cg]['csm_wocs_csc'],
                            'csm_wcs_ctwocc': cncgsm[cg]['csm_wcs_ctwocc'],
                            'csm_wcs_ctwcc': cncgsm[cg]['csm_wcs_ctwcc'],
                            'csm_wcs_csc': cncgsm[cg]['csm_wcs_csc'],}
                        ce.add_coord_geom(cg, cncgsm[cg]['csm'],
                                          algo=cncgsm[cg]['algo'],
                                          permutation=cncgsm[cg]['indices'],
                                          local2perfect_map=cncgsm[cg][
                                              'local2perfect_map'],
                                          perfect2local_map=cncgsm[cg][
                                              'perfect2local_map'],
                                          detailed_voronoi_index={'cn': cn,
                                                                  'index': i_nlist},
                                          other_symmetry_measures=other_csms
                                          )
                    ce_dict[cn].append(ce)
            t2 = time.clock()
            logging.info('    ... computed in {:.2f} seconds'.format(t2 - t1))
            ce_list.append(ce_dict)
        tse2 = time.clock()
        logging.info('Structure environments computed in {:.2f} seconds'.format(
            tse2 - tse1))
        return StructureEnvironments(self.detailed_voronoi, self.valences,
                                     self.sites_map, self.equivalent_sites,
                                     ce_list, self.structure)

    def setup_local_geometry(self, isite, coords):
        """
        Sets up the AbstractGeometry for the local geometry of site with index isite.
        :param isite: Index of the site for which the local geometry has to be set up
        :param coords: The coordinates of the (local) neighbors
        """
        self.local_geometry = AbstractGeometry(
            central_site=self.structure.cart_coords[isite],
            bare_coords=coords,
            centering_type=self.centering_type,
            include_central_site_in_centroid=
            self.include_central_site_in_centroid)

    def setup_test_perfect_environment(self, symbol, randomness=False,
                                       max_random_dist=0.1,
                                       symbol_type='mp_symbol',
                                       indices='RANDOM',
                                       random_translation=False,
                                       random_rotation=False,
                                       random_scale=False):
        if symbol_type == 'IUPAC':
            cg = self.cg.get_geometry_from_IUPAC_symbol(symbol)
        elif symbol_type == 'MP' or symbol_type == 'mp_symbol':
            cg = self.cg.get_geometry_from_mp_symbol(symbol)
        else:
            raise ValueError('Wrong mp_symbol to setup coordination geometry')
        neighb_coords = []
        if randomness:
            rv = np.random.random_sample(3)
            while norm(rv) > 1.0:
                rv = np.random.random_sample(3)
            coords = [np.zeros(3, np.float) + max_random_dist * rv]
            for pp in cg.points:
                rv = np.random.random_sample(3)
                while norm(rv) > 1.0:
                    rv = np.random.random_sample(3)
                neighb_coords.append(np.array(pp) + max_random_dist * rv)
        else:
            coords = [np.zeros(3, np.float)]
            for pp in cg.points:
                neighb_coords.append(np.array(pp))
        if indices == 'RANDOM':
            shuffle(neighb_coords)
        elif indices == 'ORDERED':
            pass
        else:
            neighb_coords = [neighb_coords[ii] for ii in indices]

        # Scaling the test environment
        if random_scale == 'RANDOM':
            scale = 0.1*np.random.random_sample() + 0.95
        elif random_scale == 'NONE':
            scale = 1.0
        else:
            scale = random_scale
        coords = [scale * cc for cc in coords]
        neighb_coords = [scale * cc for cc in neighb_coords]

        # Rotating the test environment
        if random_rotation == 'RANDOM':
            uu = np.random.random_sample(3) + 0.1
            uu = uu / norm(uu)
            theta = np.pi * np.random.random_sample()
            cc = np.cos(theta)
            ss = np.sin(theta)
            ux = uu[0]
            uy = uu[1]
            uz = uu[2]
            RR = np.matrix([[ux*ux+(1.0-ux*ux)*cc, ux*uy*(1.0-cc)-uz*ss, ux*uz*(1.0-cc)+uy*ss],
                            [ux*uy*(1.0-cc)+uz*ss, uy*uy+(1.0-uy*uy)*cc, uy*uz*(1.0-cc)-ux*ss],
                            [ux*uz*(1.0-cc)-uy*ss, uy*uz*(1.0-cc)+ux*ss, uz*uz+(1.0-uz*uz)*cc]])
        elif random_rotation == 'NONE':
            RR = [[1.0, 0.0, 0.0],
                  [0.0, 1.0, 0.0],
                  [0.0, 0.0, 1.0]]
        else:
            RR = random_rotation
        newcoords = []
        for cc in coords:
            newcc = RR * np.matrix(cc).T
            newcoords.append(newcc.getA1())
        coords = newcoords
        newcoords = []
        for cc in neighb_coords:
            newcc = RR * np.matrix(cc).T
            newcoords.append(newcc.getA1())
        neighb_coords = newcoords

        # Translating the test environment
        if random_translation == 'RANDOM':
            translation = 10.0 * (2.0*np.random.random_sample(3)-1.0)
        elif random_translation == 'NONE':
            translation = np.zeros(3, np.float)
        else:
            translation = random_translation
        coords = [cc + translation for cc in coords]
        neighb_coords = [cc + translation for cc in neighb_coords]

        coords.extend(neighb_coords)
        myspecies = ["O"] * (len(coords))
        myspecies[0] = "Cu"

        amin = np.min([cc[0] for cc in coords])
        amax = np.max([cc[0] for cc in coords])
        bmin = np.min([cc[1] for cc in coords])
        bmax = np.max([cc[1] for cc in coords])
        cmin = np.min([cc[2] for cc in coords])
        cmax = np.max([cc[2] for cc in coords])

        factor = 5.0
        aa = factor * max([amax - amin, bmax - bmin, cmax - cmin])

        lattice = Lattice.cubic(a=aa)
        structure = Structure(lattice=lattice, species=myspecies, coords=coords,
                              to_unit_cell=False, coords_are_cartesian=True)

        self.setup_structure(structure=structure)
        self.setup_local_geometry(isite=0, coords=neighb_coords)
        self.perfect_geometry = AbstractGeometry.from_cg(cg=cg)

    def setup_random_structure(self, coordination):
        """
        Sets up a purely random structure with a given coordination.
        :param coordination: coordination number for the random structure
        """
        aa = 0.4
        bb = -0.2
        coords = list()
        for ii in range(coordination + 1):
            coords.append(aa * np.random.random_sample(3, ) + bb)
        self.set_structure(
            lattice=np.array([[10, 0, 0], [0, 10, 0], [0, 0, 10]], np.float),
            species=["Si"] * (coordination + 1),
            coords=coords,
            coords_are_cartesian=False)
        self.setup_random_indices_local_geometry(coordination)

    def setup_random_indices_local_geometry(self, coordination):
        """
        Sets up random indices for the local geometry, for testing purposes
        :param coordination: coordination of the local geometry
        """
        self.icentral_site = 0
        self.indices = list(range(1, coordination + 1))
        np.random.shuffle(self.indices)

    def setup_ordered_indices_local_geometry(self, coordination):
        """
        Sets up ordered indices for the local geometry, for testing purposes
        :param coordination: coordination of the local geometry
        """
        self.icentral_site = 0
        self.indices = list(range(1, coordination + 1))

    def setup_explicit_indices_local_geometry(self, explicit_indices):
        """
        Sets up explicit indices for the local geometry, for testing purposes
        :param explicit_indices: explicit indices for the neighbors (set of numbers
        from 0 to CN-1 in a given order)
        """
        self.icentral_site = 0
        self.indices = [ii + 1 for ii in explicit_indices]

    def get_coordination_symmetry_measures(self, only_minimum=True,
                                           all_csms=True):
        """
        Returns the continuous symmetry measures of the current local geometry in a dictionary.
        :return: the continuous symmetry measures of the current local geometry in a dictionary.
        """
        test_geometries = self.cg.get_implemented_geometries(
            len(self.local_geometry.coords))
        result_dict = {}
        for geometry in test_geometries:
            self.perfect_geometry = AbstractGeometry.from_cg(cg=geometry,
                                                             centering_type=self.centering_type,
                                                             include_central_site_in_centroid=
                                                             self.include_central_site_in_centroid)
            points_perfect = self.perfect_geometry.points_wocs_ctwocc()
            cgsm = self.coordination_geometry_symmetry_measures(geometry,
                                                                points_perfect=points_perfect)
            result, permutations, algos, local2perfect_maps, perfect2local_maps = cgsm
            if only_minimum:
                if len(result) > 0:
                    imin = np.argmin(result)
                    if geometry.algorithms is not None:
                        algo = algos[imin]
                    else:
                        algo = algos
                    result_dict[geometry.mp_symbol] = {'csm': result[imin],
                                                       'indices': permutations[
                                                           imin],
                                                       'algo': algo,
                                                       'local2perfect_map':
                                                           local2perfect_maps[
                                                               imin],
                                                       'perfect2local_map':
                                                           perfect2local_maps[
                                                               imin]}
                    if all_csms:
                        permutation = permutations[imin]
                        # Without central site, centered on the centroid (centroid does not include the central site)
                        result_dict[geometry.mp_symbol]['csm_wocs_ctwocc'] = \
                        result[imin]
                        # Without central site, centered on the centroid (centroid includes the central site)
                        pdist = self.local_geometry.points_wocs_ctwcc(
                            permutation=permutation)
                        pperf = self.perfect_geometry.points_wocs_ctwcc()
                        csm = symmetry_measure(points_distorted=pdist,
                                               points_perfect=pperf)
                        result_dict[geometry.mp_symbol]['csm_wocs_ctwcc'] = csm
                        # Without central site, centered on the central site
                        pdist = self.local_geometry.points_wocs_csc(
                            permutation=permutation)
                        pperf = self.perfect_geometry.points_wocs_csc()
                        csm = symmetry_measure(points_distorted=pdist,
                                               points_perfect=pperf)
                        result_dict[geometry.mp_symbol]['csm_wocs_csc'] = csm
                        # With central site, centered on the centroid (centroid does not include the central site)
                        pdist = self.local_geometry.points_wcs_ctwocc(
                            permutation=permutation)
                        pperf = self.perfect_geometry.points_wcs_ctwocc()
                        csm = symmetry_measure(points_distorted=pdist,
                                               points_perfect=pperf)
                        result_dict[geometry.mp_symbol]['csm_wcs_ctwocc'] = csm
                        # With central site, centered on the centroid (centroid includes the central site)
                        pdist = self.local_geometry.points_wcs_ctwcc(
                            permutation=permutation)
                        pperf = self.perfect_geometry.points_wcs_ctwcc()
                        csm = symmetry_measure(points_distorted=pdist,
                                               points_perfect=pperf)
                        result_dict[geometry.mp_symbol]['csm_wcs_ctwcc'] = csm
                        # With central site, centered on the central site

                        pdist = self.local_geometry.points_wcs_csc(
                            permutation=permutation)
                        pperf = self.perfect_geometry.points_wcs_csc()

                        csm = symmetry_measure(points_distorted=pdist,
                                               points_perfect=pperf)
                        result_dict[geometry.mp_symbol]['csm_wcs_csc'] = csm

            else:
                result_dict[geometry.mp_symbol] = {'csm': result,
                                                   'indices': permutations,
                                                   'algo': algos,
                                                   'local2perfect_map': local2perfect_maps,
                                                   'perfect2local_map': perfect2local_maps}
        return result_dict

    def coordination_geometry_symmetry_measures(self, coordination_geometry,
                                                tested_permutations=False,
                                                points_perfect=None):
        """
        Returns the symmetry measures of a given coordination_geometry for a set of permutations depending on
        the permutation setup. Depending on the parameters of the LocalGeometryFinder and on the coordination
         geometry, different methods are called.
        :param coordination_geometry: Coordination geometry for which the symmetry measures are looked for
        :return: the symmetry measures of a given coordination_geometry for a set of permutations
        :raise: NotImplementedError if the permutation_setup does not exists
        """
        if tested_permutations:
            tested_permutations = set()
        if self.permutations_safe_override:
            return self.coordination_geometry_symmetry_measures_safe_override(
                coordination_geometry)
        csms = []
        permutations = []
        algos = []
        local2perfect_maps = []
        perfect2local_maps = []
        for algo in coordination_geometry.algorithms:
            if algo.algorithm_type == EXPLICIT_PERMUTATIONS:
                return self.coordination_geometry_symmetry_measures_standard(
                    coordination_geometry, algo,
                    points_perfect=points_perfect)
            if algo.algorithm_type == SEPARATION_PLANE:
                cgsm = self.coordination_geometry_symmetry_measures_separation_plane(
                    coordination_geometry,
                    algo,
                    tested_permutations=tested_permutations,
                    points_perfect=points_perfect)
                csm, perm, algo, local2perfect_map, perfect2local_map = cgsm

                csms.extend(csm)
                permutations.extend(perm)
                algos.extend(algo)
                local2perfect_maps.extend(local2perfect_map)
                perfect2local_maps.extend(perfect2local_map)
        return csms, permutations, algos, local2perfect_maps, perfect2local_maps

    def coordination_geometry_symmetry_measures_safe_override(self,
                                                              coordination_geometry):
        """
        Returns the symmetry measures for a set of permutations (whose setup depends on the coordination geometry)
        for the coordination geometry "coordination_geometry". Standard implementatison looking for the symmetry
        measures of each permutation

        :param coordination_geometry: The coordination geometry to be investigated
        :return: The symmetry measures for the given coordination geometry for each permutation investigated
        """
        permutations_symmetry_measures = np.zeros(
            coordination_geometry.number_of_permutations, np.float)
        permutations = list()
        iperm = 0
        for perm in itertools.permutations(
                list(range(coordination_geometry.coordination_number))):
            permutations.append(np.argsort(perm))
            bare_coords = list()
            for ii in perm:
                bare_coords.append(coordination_geometry.points[ii])
            perfect_geometry = AbstractGeometry(
                central_site=coordination_geometry.get_central_site(),
                bare_coords=bare_coords,
                centering_type=self.centering_type,
                include_central_site_in_centroid=
                self.include_central_site_in_centroid)
            permutations_symmetry_measures[
                iperm] = self.symmetry_measure_newpmg(self.local_geometry,
                                                      perfect_geometry)
            iperm += 1
        return permutations_symmetry_measures, permutations, 'SAFE'

    def coordination_geometry_symmetry_measures_standard(self,
                                                         coordination_geometry,
                                                         algo,
                                                         points_perfect=None):
        """
        Returns the symmetry measures for a set of permutations (whose setup depends on the coordination geometry)
        for the coordination geometry "coordination_geometry". Standard implementation looking for the symmetry
        measures of each permutation

        :param coordination_geometry: The coordination geometry to be investigated
        :return: The symmetry measures for the given coordination geometry for each permutation investigated
        """
        permutations_symmetry_measures = np.zeros(len(algo.permutations),
                                                  np.float)
        permutations = list()
        algos = list()
        local2perfect_maps = list()
        perfect2local_maps = list()
        for iperm, perm in enumerate(algo.permutations):

            local2perfect_map = {}
            perfect2local_map = {}
            permutations.append(perm)
            for iperfect, ii in enumerate(perm):
                perfect2local_map[iperfect] = ii
                local2perfect_map[ii] = iperfect
            local2perfect_maps.append(local2perfect_map)
            perfect2local_maps.append(perfect2local_map)

            points_distorted = self.local_geometry.points_wocs_ctwocc(
                permutation=perm)

            csm = symmetry_measure(points_distorted=points_distorted,
                                   points_perfect=points_perfect)

            permutations_symmetry_measures[iperm] = csm
            algos.append(str(algo))
        return permutations_symmetry_measures, permutations, algos, local2perfect_maps, perfect2local_maps

    def coordination_geometry_symmetry_measures_separation_plane(self,
                                                                 coordination_geometry,
                                                                 separation_plane_algo,
                                                                 testing=False,
                                                                 tested_permutations=False,
                                                                 points_perfect=None):

        """
        Returns the symmetry measures of the given coordination geometry "coordination_geometry" using separation
        planes to reduce the complexity of the system. Caller to the refined 2POINTS, 3POINTS and other ...
        :param coordination_geometry: The coordination geometry to be investigated
        :return: The symmetry measures for the given coordination geometry for each plane and permutation investigated
        """
        permutations = list()
        permutations_symmetry_measures = list()
        plane_separations = list()
        algos = list()
        perfect2local_maps = list()
        local2perfect_maps = list()
        if testing:
            separation_permutations = list()
        nplanes = 0
        for npoints in range(separation_plane_algo.minimum_number_of_points,
                             min(separation_plane_algo.maximum_number_of_points,
                                 4) + 1):
            for points_combination in itertools.combinations(
                    self.local_geometry.coords, npoints):
                if npoints == 2:
                    if collinear(points_combination[0], points_combination[1],
                                 self.local_geometry.central_site,
                                 tolerance=0.25):
                        continue
                    plane = Plane.from_3points(points_combination[0],
                                               points_combination[1],
                                               self.local_geometry.central_site)
                elif npoints == 3:
                    if collinear(points_combination[0], points_combination[1],
                                 points_combination[2], tolerance=0.25):
                        continue
                    plane = Plane.from_3points(points_combination[0],
                                               points_combination[1],
                                               points_combination[2])
                elif npoints > 3:
                    plane = Plane.from_npoints(points_combination,
                                               best_fit='least_square_distance')
                else:
                    raise ValueError(
                        'Wrong number of points to initialize separation plane')
                cgsm = self._cg_csm_separation_plane(
                    coordination_geometry=coordination_geometry,
                    sepplane=separation_plane_algo,
                    local_plane=plane,
                    plane_separations=plane_separations,
                    dist_tolerances=DIST_TOLERANCES,
                    testing=testing,
                    tested_permutations=tested_permutations,
                    points_perfect=points_perfect)
                csm, perm, algo = cgsm[0], cgsm[1], cgsm[2]

                if csm is not None:
                    permutations_symmetry_measures.extend(csm)
                    permutations.extend(perm)
                    for thisperm in perm:
                        p2l = {}
                        l2p = {}
                        for i_p, pp in enumerate(thisperm):
                            p2l[i_p] = pp
                            l2p[pp] = i_p
                        perfect2local_maps.append(p2l)
                        local2perfect_maps.append(l2p)
                    algos.extend(algo)
                    if testing:
                        separation_permutations.extend(cgsm[3])
                    nplanes += 1
            if nplanes > 0:
                break
        if nplanes == 0:
            return self.coordination_geometry_symmetry_measures_fallback_random(
                coordination_geometry,
                points_perfect=points_perfect)
        if testing:
            return np.array(permutations_symmetry_measures,
                            np.float), permutations, separation_permutations
        return np.array(permutations_symmetry_measures,
                        np.float), permutations, algos, local2perfect_maps, perfect2local_maps

    def _cg_csm_separation_plane(self, coordination_geometry, sepplane,
                                 local_plane,
                                 plane_separations,
                                 dist_tolerances=DIST_TOLERANCES,
                                 testing=False, tested_permutations=False,
                                 points_perfect=None):
        argref_separation = sepplane.argsorted_ref_separation_perm
        plane_found = False
        for dist_tolerance in dist_tolerances:
            permutations = []
            permutations_symmetry_measures = []
            if testing:
                separation_permutations = []
            algo = 'NOT_FOUND'
            separation = local_plane.indices_separate(
                self.local_geometry._coords, dist_tolerance)
            # Do not consider planes leading to the same separation indices
            separation = sort_separation(separation)

            if separation_in_list(separation, plane_separations):
                continue
            # Do not consider a separation which does not follow the reference separation of the perfect
            # coordination geometry
            if len(separation[1]) != len(sepplane.plane_points):
                continue
            if len(separation[0]) == len(sepplane.point_groups[0]):
                this_separation = separation
                plane_separations.append(this_separation)
            elif len(separation[0]) == len(sepplane.point_groups[1]):
                this_separation = [list(separation[2]), list(separation[1]),
                                   list(separation[0])]
                plane_separations.append(this_separation)
            else:
                continue

            if sepplane.ordered_plane:
                inp = [pp for ip, pp in enumerate(self.local_geometry._coords)
                       if ip in this_separation[1]]

                if sepplane.ordered_point_groups[0]:
                    pp_s0 = [pp for ip, pp in
                             enumerate(self.local_geometry._coords) if
                             ip in this_separation[0]]
                    ordind_s0 = local_plane.project_and_to2dim_ordered_indices(
                        pp_s0)
                    sep0 = [this_separation[0][ii] for ii in ordind_s0]
                else:
                    sep0 = list(this_separation[0])
                if sepplane.ordered_point_groups[1]:
                    pp_s2 = [pp for ip, pp in
                             enumerate(self.local_geometry._coords) if
                             ip in this_separation[2]]
                    ordind_s2 = local_plane.project_and_to2dim_ordered_indices(
                        pp_s2)
                    sep2 = [this_separation[2][ii] for ii in ordind_s2]
                else:
                    sep2 = list(this_separation[2])
                separation_perm = list(sep0)
                ordind = local_plane.project_and_to2dim_ordered_indices(inp)
                separation_perm.extend(
                    [this_separation[1][ii] for ii in ordind])
                algo = 'SEPARATION_PLANE_2POINTS_ORDERED'
                separation_perm.extend(sep2)
            else:
                separation_perm = list(this_separation[0])
                separation_perm.extend(this_separation[1])
                algo = 'SEPARATION_PLANE_2POINTS'
                separation_perm.extend(this_separation[2])
            if self.plane_safe_permutations:
                sep_perms = sepplane.safe_separation_permutations(
                    ordered_plane=sepplane.ordered_plane,
                    ordered_point_groups=sepplane.ordered_point_groups)
            else:
                sep_perms = sepplane.permutations

            plane_found = True

            for i_sep_perm, sep_perm in enumerate(sep_perms):
                perm1 = [separation_perm[ii] for ii in sep_perm]
                pp = [perm1[ii] for ii in argref_separation]
                # Skip permutations that have already been performed
                if tested_permutations != False and coordination_geometry.equivalent_indices is not None:
                    tuple_ref_perm = coordination_geometry.ref_permutation(pp)
                    if tuple_ref_perm in tested_permutations:
                        continue
                    tested_permutations.add(tuple_ref_perm)

                permutations.append(pp)
                if testing:
                    separation_permutations.append(sep_perm)

                points_distorted = self.local_geometry.points_wocs_ctwocc(
                    permutation=pp)

                csm = symmetry_measure(points_distorted=points_distorted,
                                       points_perfect=points_perfect)

                permutations_symmetry_measures.append(csm)
            if plane_found:
                break
        if len(permutations_symmetry_measures) > 0:
            if testing:
                return permutations_symmetry_measures, permutations, algo, separation_permutations
            else:
                return permutations_symmetry_measures, permutations, [
                    sepplane.algorithm_type] * len(permutations)
        else:
            if plane_found:
                return permutations_symmetry_measures, permutations, []
            else:
                return None, None, None

    def coordination_geometry_symmetry_measures_fallback_random(self,
                                                                coordination_geometry,
                                                                NRANDOM=200,
                                                                points_perfect=None):
        """
        Returns the symmetry measures for a random set of permutations for the coordination geometry
        "coordination_geometry". Fallback implementation for the plane separation algorithms measures
        of each permutation

        :param coordination_geometry: The coordination geometry to be investigated
        :param NRANDOM: Number of random permutations to be tested
        :return: The symmetry measures for the given coordination geometry for each permutation investigated
        """
        permutations_symmetry_measures = np.zeros(NRANDOM, np.float)
        permutations = list()
        algos = list()
        perfect2local_maps = list()
        local2perfect_maps = list()
        for iperm in range(NRANDOM):
            perm = np.random.permutation(
                coordination_geometry.coordination_number)
            permutations.append(perm)
            p2l = {}
            l2p = {}
            for i_p, pp in enumerate(perm):
                p2l[i_p] = pp
                l2p[pp] = i_p
            perfect2local_maps.append(p2l)
            local2perfect_maps.append(l2p)

            points_distorted = self.local_geometry.points_wocs_ctwocc(
                permutation=perm)
            csm = symmetry_measure(points_distorted=points_distorted,
                                   points_perfect=points_perfect)
            permutations_symmetry_measures[iperm] = csm
            algos.append('APPROXIMATE_FALLBACK')
        return permutations_symmetry_measures, permutations, algos, local2perfect_maps, perfect2local_maps



# coding: utf-8
# Copyright (c) Pymatgen Development Team.
# Distributed under the terms of the MIT License.

from __future__ import division, print_function, unicode_literals, absolute_import

"""
This module implements classes for generating Lammps input.

For the ease of management we divide LAMMPS input into 2 files:

    1.Data file: All structure related settings such as the atomic positions,
            bonds, angles, dihedrals, corresponding parametrizations etc are
            set in the data file.

    2. Control file: This the main input file that should be fed to the
            lammps binary. The main input file consists of the path to the
            afore-mentioned data file and the job control parameters such as
            the ensemble type(NVT, NPT etc), max number of iterations etc.
"""

import six
import json   
import os
from functools import partial
from collections import OrderedDict

from monty.json import MSONable, MontyDecoder

from pymatgen.io.lammps.data import LammpsData, LammpsForceFieldData

__author__ = "Kiran Mathew"
__email__ = "kmathew@lbl.gov"
__credits__ = "Navnidhi Rajput"

MODULE_DIR = os.path.dirname(os.path.abspath(__file__))


class DictLammpsInput(MSONable):
    """
    Implementation of LammpsInputSet that is initialized from a dict
    settings. It is typically used by other LammpsInputSets for
    initialization from json or yaml source files.

    Args:
        name (str): A name for the input set.
        config_dict (dict): The config dictionary to use.
        lammps_data (LammpsData): LammpsData object
        data_filename (str): name of the the lammps data file
        user_lammps_settings (dict): User lammps settings. This allows a user
            to override lammps settings, e.g., setting a different force field
            or bond type.
    """

    def __init__(self, name, config_dict, lammps_data=None,
                 data_filename="in.data", user_lammps_settings={}):
        self.name = name
        self.lines = []
        self.config_dict = config_dict
        self.lammps_data = lammps_data
        self.data_filename = data_filename
        self.config_dict["read_data"] = data_filename
        self.user_lammps_settings = user_lammps_settings
        if self.user_lammps_settings:
            self.config_dict.update(self.user_lammps_settings)

    def __str__(self):
        """
        string representation of the lammps input file with the
        control parameters
        """
        lines = ""
        for k1, v1 in self.config_dict.items():
            if isinstance(v1, dict):
                v1 = v1.values()
            if isinstance(v1, list):
                for x in v1:
                    lines = "".join([lines, "{} ".format(k1)])
                    lines = "".join([lines, str(x), os.linesep])
            else:
                lines = "".join([lines, "{} ".format(k1)])
                lines = "".join([lines, " {}{}".format(str(v1), os.linesep)])
        return lines

    def write_input(self, filename, data_filename=None):
        """
        Get the string representation of the main input file and write it.
        Also writes the data file if the lammps_data attribute is set.

        Args:
            filename (string): name of the input file
            data_filename (string): override the data file name with this
        """
        if data_filename:
            self.config_dict["read_data"] = data_filename
            self.data_filename = data_filename
        # write the main input file
        with open(filename, 'w') as f:
            f.write(self.__str__())
        # write the data file if present
        if self.lammps_data:
            print("Data file: {}".format(self.data_filename))
            self.lammps_data.write_data_file(filename=self.data_filename)

    @staticmethod
    def from_file(name, filename, lammps_data=None, data_filename="in.data",
                  user_lammps_settings={}, is_forcefield=False):
        """
        Read in the input settings from json file as ordereddict.
        Note: with monty.serialization.loadfn the order of paramters in the
        json file is not preserved

        Args:
            filename (string): name of the file with the lamps control
                paramters
            lammps_data (string/LammpsData/LammpsForceFieldData): path to the
                data file or an appropriate object
            data_filename (string): name of the the lammps data file
            user_lammps_settings (dict): User lammps settings
            is_forcefield (bool): whether the data file has forcefield and
                topology info in it. This is required only if lammps_data is
                a path to the data file instead of a data object

        Returns:
            DictLammpsInput
        """
        with open(filename) as f:
            config_dict = json.load(f, object_pairs_hook=OrderedDict)
        lammps_data = lammps_data
        if isinstance(lammps_data, six.string_types):
            if is_forcefield:
                lammps_data = LammpsForceFieldData.from_file(lammps_data)
            else:
                lammps_data = LammpsData.from_file(lammps_data)
        return DictLammpsInput(name, config_dict, lammps_data=lammps_data,
                               data_filename=data_filename,
                               user_lammps_settings=user_lammps_settings)

    def as_dict(self):
        d = MSONable.as_dict(self)
        if hasattr(self, "kwargs"):
            d.update(**self.kwargs)
        d["config_dict"] = self.config_dict.items()
        return d

    @classmethod
    def from_dict(cls, d):
        decoded = {k: MontyDecoder().process_decoded(v) for k, v in d.items()
                   if k not in ["@module", "@class", "config_dict"]}
        decoded["config_dict"] = OrderedDict(d["config_dict"])
        return cls(**decoded)


# NVT
NVTLammpsInput = partial(DictLammpsInput.from_file, "NVT",
                         os.path.join(MODULE_DIR, "NVT.json"))

# NPT
NPTLammpsInput = partial(DictLammpsInput.from_file, "NPT",
                         os.path.join(MODULE_DIR, "NPT.json"))

# NPT followed by NVT
NPTNVTLammpsInput = partial(DictLammpsInput.from_file, "NPT_NVT",
                            os.path.join(MODULE_DIR, "NPT_NVT.json"))



# coding: utf-8
# Copyright (c) Pymatgen Development Team.
# Distributed under the terms of the MIT License.
#!/usr/bin/env python

from __future__ import division, unicode_literals

import warnings
warnings.warn("pymatgen.io.abinit.eos has been moved pymatgen.analysis.eos "
              "This stub will be removed in pymatgen 4.0.", DeprecationWarning)

from pymatgen.analysis.eos import *



# coding: utf-8
# Copyright (c) Pymatgen Development Team.
# Distributed under the terms of the MIT License.

from __future__ import division, print_function, unicode_literals, absolute_import

"""
This module defines classes that set the molecular topology i.e atoms, bonds,
angles and dihedrals
"""

import itertools
from pymatgen.core.bonds import CovalentBond

__author__ = 'Kiran Mathew'
__email__ = 'kmathew@lbl.gov'
__credits__ = 'Brandon Wood'


class Topology(object):
    """
    Args:
        atoms (list): map atom names to force field(ff) atom name,
            [['c', 'c1'],...]
        bonds (list): List of bonds,
            [[i,j, bond_type], ... ] where i, j are integer(starts from 1)
            atom ids in the molecules and bond_type = (ff atomname_i, ff atomname_j)
        angles (list): List of angles,
            [[i,j,k, angle_type], ... ],
            angle_type = (ff atomname_i, ff atomname_j, ff atomname_k)
        charges (list): List of charges, [0.4, 0.7, ... ]
        dihedrals (list): List of dihedrals,
            [[i,j,k,l, dihedral_type], ... ]
            dihedral_type = (ff atomname_i, ff atomname_j, ff atomname_k, ff atomname_l)
        imdihedrals (list): List of improper dihedrals,
            [['i,j,k,l, dihedral_type], ... ]
    """

    def __init__(self, atoms, bonds, angles, charges=None, dihedrals=None,
                 imdihedrals=None):
        self.atoms = atoms
        self.bonds = bonds
        self.angles = angles
        self.charges = [] if charges is None else charges
        self.dihedrals = dihedrals
        self.imdihedrals = imdihedrals

    @staticmethod
    def from_molecule(molecule, tol=0.1, ff_map="ff_map"):
        """
        Return Topology object from molecule. Charges are also set if the
        molecule has 'charge' site property.

        Args:
            molecule (Molecule)
            tol (float): Relative tolerance to test in determining the bonds
                in the molecule. Basically, the code checks if the distance
                between the sites is less than (1 + tol) * typical bond
                distances. Defaults to 0.1, i.e., 10% longer.
            ff_map (string): Ensure this site property is set for each site if
                atoms need to be mapped to its forcefield name.
                eg: Carbon atom, 'C', on different sites can be mapped to
                either 'Ce' or 'Cm' depending on how the forcefield parameters
                are set.

        Returns:
            Topology object

        """
        type_attrib = ff_map if hasattr(molecule[0], ff_map) else "specie"
        bonds = molecule.get_covalent_bonds(tol=tol)
        angles = []
        dihedrals = []
        for bond1, bond2 in itertools.combinations(bonds, 2):
            bond1_sites = [bond1.site1, bond1.site2]
            bond2_sites = [bond2.site1, bond2.site2]
            s_bond1 = set(bond1_sites)
            s_bond2 = set(bond2_sites)
            common_site = s_bond1.intersection(s_bond2)
            if common_site:
                site1 = (s_bond1 - s_bond2).pop()
                site2 = common_site.pop()
                site3 = (s_bond2 - s_bond1).pop()
                angle = [molecule.index(site1), molecule.index(site2),
                         molecule.index(site3),
                         (str(getattr(site1, type_attrib)),
                          str(getattr(site2, type_attrib)),
                          str(getattr(site3, type_attrib)))]
                angles.append(angle)
            else:
                for site1, site2 in itertools.product(bond1_sites,
                                                      bond2_sites):
                    if CovalentBond.is_bonded(site1, site2, tol=tol):
                        bond1_sites.remove(site1)
                        bond2_sites.remove(site2)
                        dihedral = bond1_sites + [site1,
                                                  site2] + bond2_sites
                        dihedral = [molecule.index(dihedral[0]),
                                    molecule.index(dihedral[1]),
                                    molecule.index(dihedral[2]),
                                    molecule.index(dihedral[3]),
                                    (str(getattr(dihedral[0], type_attrib)),
                                     str(getattr(dihedral[1], type_attrib)),
                                     str(getattr(dihedral[2], type_attrib)),
                                     str(getattr(dihedral[3], type_attrib)))]
                        dihedrals.append(dihedral)
                        break
        atoms = [[str(site.specie), str(getattr(site, type_attrib))]
                 for site in molecule]
        bonds = [[molecule.index(b.site1), molecule.index(b.site2),
                  (str(getattr(b.site1, type_attrib)),
                   str(getattr(b.site2, type_attrib)))] for b in bonds]
        charges = None
        if hasattr(molecule[0], "charge"):
            charges = [site.charge for site in molecule]
        return Topology(atoms, bonds, angles,
                        charges=charges, dihedrals=dihedrals)



#!/usr/bin/env python


__author__ = 'waroquiers'

import unittest2 as unittest
from pymatgen.analysis.chemenv.utils.func_utils import CSMFiniteRatioFunction
from pymatgen.analysis.chemenv.utils.func_utils import CSMInfiniteRatioFunction
from pymatgen.analysis.chemenv.utils.func_utils import DeltaCSMRatioFunction
import numpy as np


class FuncUtilsTest(unittest.TestCase):

    def test_CSMFiniteRatioFunction(self):
        max_csm = 8.0
        alpha = 1.0
        csm_finite_ratio = CSMFiniteRatioFunction(function='power2_decreasing_exp',
                                                  options_dict={'max_csm': max_csm,
                                                                'alpha': alpha})
        self.assertEqual(csm_finite_ratio.evaluate(0.0), 1.0)
        self.assertEqual(csm_finite_ratio.evaluate(2.0), 0.43807544047766522)
        self.assertEqual(csm_finite_ratio.evaluate(4.0), 0.15163266492815836)
        self.assertEqual(csm_finite_ratio.evaluate(8.0), 0.0)
        self.assertEqual(csm_finite_ratio.evaluate(9.0), 0.0)

        max_csm = 8.0
        alpha = 2.0
        csm_finite_ratio = CSMFiniteRatioFunction(function='power2_decreasing_exp',
                                                  options_dict={'max_csm': max_csm,
                                                                'alpha': alpha})
        self.assertEqual(csm_finite_ratio.evaluate(0.0), 1.0)
        self.assertEqual(csm_finite_ratio.evaluate(4.0), 0.091969860292860584)
        self.assertEqual(csm_finite_ratio.evaluate(8.0), 0.0)
        self.assertEqual(csm_finite_ratio.evaluate(9.0), 0.0)

        max_csm = 4.0
        alpha = 1.0
        csm_finite_ratio = CSMFiniteRatioFunction(function='power2_decreasing_exp',
                                                  options_dict={'max_csm': max_csm,
                                                                'alpha': alpha})
        self.assertEqual(csm_finite_ratio.evaluate(0.0), 1.0)
        self.assertEqual(csm_finite_ratio.evaluate(1.0), 0.43807544047766522)
        self.assertEqual(csm_finite_ratio.evaluate(2.0), 0.15163266492815836)
        self.assertEqual(csm_finite_ratio.evaluate(4.0), 0.0)
        self.assertEqual(csm_finite_ratio.evaluate(4.5), 0.0)

        self.assertRaises(ValueError, CSMFiniteRatioFunction,
                          function='powern_decreasing',
                          options_dict={'max_csm': max_csm,
                                        'nn': 2})

    def test_CSMInfiniteRatioFunction(self):
        max_csm = 8.0
        self.assertRaises(ValueError, CSMInfiniteRatioFunction,
                          function='power2_inverse_decreasing',
                          options_dict={'max_csm': max_csm,
                                        'nn': 2})

        self.assertRaises(ValueError, CSMInfiniteRatioFunction,
                          function='power2_tangent_decreasing',
                          options_dict={'max_csm': max_csm})

        csm_infinite_ratio = CSMInfiniteRatioFunction(function='power2_inverse_decreasing',
                                                      options_dict={'max_csm': max_csm})
        # csm_infinite_ratio = CSMInfiniteRatioFunction(function='power2_inverse_decreasing')
        self.assertEqual(csm_infinite_ratio.evaluate(0.0), np.inf)
        self.assertEqual(csm_infinite_ratio.evaluate(2.0), 2.25)
        self.assertEqual(csm_infinite_ratio.evaluate(4.0), 0.5)
        self.assertEqual(csm_infinite_ratio.evaluate(8.0), 0.0)
        self.assertEqual(csm_infinite_ratio.evaluate(9.0), 0.0)

        csm_infinite_ratio = CSMInfiniteRatioFunction(function='power2_inverse_power2_decreasing',
                                                      options_dict={'max_csm': max_csm})
        self.assertEqual(csm_infinite_ratio.evaluate(0.0), np.inf)
        self.assertEqual(csm_infinite_ratio.evaluate(2.0), 9.0)
        self.assertEqual(csm_infinite_ratio.evaluate(4.0), 1.0)
        self.assertEqual(csm_infinite_ratio.evaluate(8.0), 0.0)
        self.assertEqual(csm_infinite_ratio.evaluate(9.0), 0.0)

        max_csm = 12.0
        csm_infinite_ratio = CSMInfiniteRatioFunction(function='power2_inverse_power2_decreasing',
                                                      options_dict={'max_csm': max_csm})

        self.assertEqual(csm_infinite_ratio.evaluate(0.0), np.inf)
        self.assertEqual(csm_infinite_ratio.evaluate(3.0), 9.0)
        self.assertEqual(csm_infinite_ratio.evaluate(6.0), 1.0)
        self.assertEqual(csm_infinite_ratio.evaluate(12.0), 0.0)
        self.assertEqual(csm_infinite_ratio.evaluate(20.0), 0.0)

    def test_DeltaCSMRatioFunction(self):
        self.assertRaises(ValueError, DeltaCSMRatioFunction,
                          function='smoothstep',
                          options_dict={})
        self.assertRaises(ValueError, DeltaCSMRatioFunction,
                          function='smootherstep',
                          options_dict={})
        self.assertRaises(ValueError, DeltaCSMRatioFunction,
                          function='smootherstep',
                          options_dict={'delta_csm_min': 1.0})
        self.assertRaises(ValueError, DeltaCSMRatioFunction,
                          function='smootherstep',
                          options_dict={'delta_csm_max': 1.0})

        delta_csm_ratio_function = DeltaCSMRatioFunction(function='smootherstep',
                                                         options_dict={'delta_csm_min': 1.0, 'delta_csm_max': 4.0})
        self.assertEqual(delta_csm_ratio_function.evaluate(0.0), 0.0)
        self.assertEqual(delta_csm_ratio_function.evaluate(1.0), 0.0)
        self.assertEqual(delta_csm_ratio_function.evaluate(2.5), 0.5)
        self.assertEqual(delta_csm_ratio_function.evaluate(4.0), 1.0)
        self.assertEqual(delta_csm_ratio_function.evaluate(5.0), 1.0)

        delta_csm_ratio_function = DeltaCSMRatioFunction(function='smootherstep',
                                                         options_dict={'delta_csm_min': 2.0, 'delta_csm_max': 8.0})
        self.assertEqual(delta_csm_ratio_function.evaluate(0.0), 0.0)
        self.assertEqual(delta_csm_ratio_function.evaluate(2.0), 0.0)
        self.assertEqual(delta_csm_ratio_function.evaluate(5.0), 0.5)
        self.assertEqual(delta_csm_ratio_function.evaluate(8.0), 1.0)
        self.assertEqual(delta_csm_ratio_function.evaluate(12.0), 1.0)

if __name__ == "__main__":
    unittest.main()